"""Code for Gramian Angular Field."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

import numpy as np
from math import ceil
from numba import njit, prange
from sklearn.base import BaseEstimator
from sklearn.utils.validation import check_array
from ..approximation import PiecewiseAggregateApproximation
from ..base import UnivariateTransformerMixin
from ..preprocessing import MinMaxScaler


@njit()
def _gasf(X_cos, X_sin, n_samples, image_size):
    X_gasf = np.empty((n_samples, image_size, image_size))
    for i in prange(n_samples):
        X_gasf[i] = np.outer(X_cos[i], X_cos[i]) - np.outer(X_sin[i], X_sin[i])
    return X_gasf


@njit()
def _gadf(X_cos, X_sin, n_samples, image_size):
    X_gadf = np.empty((n_samples, image_size, image_size))
    for i in prange(n_samples):
        X_gadf[i] = np.outer(X_sin[i], X_cos[i]) - np.outer(X_cos[i], X_sin[i])
    return X_gadf


class GramianAngularField(BaseEstimator, UnivariateTransformerMixin):
    """Gramian Angular Field.

    Parameters
    ----------
    image_size : int or float (default = 1.)
        Shape of the output images. If float, it represents a percentage
        of the size of each time series and must be between 0 and 1. Output
        images are square, thus providing the size of one dimension is enough.

    sample_range : None or tuple (min, max) (default = (-1, 1))
        Desired range of transformed data. If None, no scaling is performed
        and all the values of the input data must be between -1 and 1.
        If tuple, each sample is scaled between min and max; min must be
        greater than or equal to -1 and max must be lower than or equal to 1.

    method : 'summation' or 'difference' (default = 'summation')
        Type of Gramian Angular Field. 's' can be used for 'summation'
        and 'd' for 'difference'.

    overlapping : bool (default = False)
        If True, reduce the size of each time series using PAA with possible
        overlapping windows.

    flatten : bool (default = False)
        If True, images are flattened to be one-dimensional.

    References
    ----------
    .. [1] Z. Wang and T. Oates, "Encoding Time Series as Images for Visual
           Inspection and Classification Using Tiled Convolutional Neural
           Networks." AAAI Workshop (2015).

    Examples
    --------
    >>> from pyts.datasets import load_gunpoint
    >>> from pyts.image import GramianAngularField
    >>> X, _, _, _ = load_gunpoint(return_X_y=True)
    >>> transformer = GramianAngularField()
    >>> X_new = transformer.transform(X)
    >>> X_new.shape
    (50, 150, 150)

    """

    def __init__(self, image_size=1., sample_range=(-1, 1),
                 method='summation', overlapping=False, flatten=False):
        self.image_size = image_size
        self.sample_range = sample_range
        self.method = method
        self.overlapping = overlapping
        self.flatten = flatten

    def fit(self, X=None, y=None):
        """Pass.

        Parameters
        ----------
        X
            Ignored
        y
            Ignored

        Returns
        -------
        self : object

        """
        return self

    def transform(self, X):
        """Transform each time series into a GAF image.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_timestamps)

        Returns
        -------
        X_new : array-like, shape = (n_samples, image_size, image_size)
            Transformed data. If ``flatten=True``, the shape is
            `(n_samples, image_size * image_size)`.

        """
        X = check_array(X)
        n_samples, n_timestamps = X.shape
        image_size = self._check_params(n_timestamps)

        paa = PiecewiseAggregateApproximation(
            window_size=None, output_size=image_size,
            overlapping=self.overlapping
        )
        X_paa = paa.fit_transform(X)
        if self.sample_range is None:
            X_min, X_max = np.min(X_paa), np.max(X_paa)
            if (X_min < -1) or (X_max > 1):
                raise ValueError("If 'sample_range' is None, all the values "
                                 "of X must be between -1 and 1.")
            X_cos = X_paa
        else:
            scaler = MinMaxScaler(sample_range=self.sample_range)
            X_cos = scaler.fit_transform(X_paa)
        X_sin = np.sqrt(np.clip(1 - X_cos ** 2, 0, 1))
        if self.method in ['s', 'summation']:
            X_new = _gasf(X_cos, X_sin, n_samples, image_size)
        else:
            X_new = _gadf(X_cos, X_sin, n_samples, image_size)

        if self.flatten:
            return X_new.reshape(n_samples, -1)
        return X_new

    def _check_params(self, n_timestamps):
        if not isinstance(self.image_size,
                          (int, np.integer, float, np.floating)):
            raise TypeError("'image_size' must be an integer or a float.")
        if isinstance(self.image_size, (int, np.integer)):
            if self.image_size < 1 or self.image_size > n_timestamps:
                raise ValueError(
                    "If 'image_size' is an integer, it must be greater "
                    "than or equal to 1 and lower than or equal to "
                    "n_timestamps (got {0}).".format(self.image_size)
                )
            image_size = self.image_size
        else:
            if not (0 < self.image_size <= 1.):
                raise ValueError(
                    "If 'image_size' is a float, it must be greater "
                    "than 0 and lower than or equal to 1 (got {0})."
                    .format(self.image_size)
                )
            image_size = ceil(self.image_size * n_timestamps)
        if not ((self.sample_range is None)
                or (isinstance(self.sample_range, tuple))):
            raise TypeError("'sample_range' must be None or a tuple.")
        if isinstance(self.sample_range, tuple):
            if len(self.sample_range) != 2:
                raise ValueError("If 'sample_range' is a tuple, its length "
                                 "must be equal to 2.")
            if not -1 <= self.sample_range[0] < self.sample_range[1] <= 1:
                raise ValueError(
                    "If 'sample_range' is a tuple, it must satisfy "
                    "-1 <= sample_range[0] < sample_range[1] <= 1."
                )
        if self.method not in ['s', 'd', 'summation', 'difference']:
            raise ValueError("'method' must be either 'summation', 's', "
                             "'difference' or 'd'.")
        return image_size



"""Testing for Time Series Forest."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

import numpy as np
import pytest
import re
from pyts.classification import TimeSeriesForest
from pyts.classification.time_series_forest import WindowFeatureExtractor


n_samples, n_timestamps = 5, 20
X_arange = np.arange(n_samples * n_timestamps).reshape(n_samples, n_timestamps)
X_ones = np.ones((n_samples, n_timestamps))
y = [0, 0, 1, 1, 2]


@pytest.mark.parametrize(
    'params, error, err_msg',
    [({'n_windows': '4'}, TypeError,
      "'n_windows' must be an integer or a float."),

     ({'min_window_size': None}, TypeError,
      "'min_window_size' must be an integer or a float."),

     ({'n_windows': 0}, ValueError,
      "If 'n_windows' is an integer, it must be positive (got 0)."),

     ({'n_windows': -3}, ValueError,
      "If 'n_windows' is an integer, it must be positive (got -3)."),

     ({'n_windows': -0.5}, ValueError,
      "If 'n_windows' is a float, it must be greater than 0 (got -0.5)."),

     ({'n_windows': -2.}, ValueError,
      "If 'n_windows' is a float, it must be greater than 0 (got -2.0)."),

     ({'min_window_size': 0}, ValueError,
      "If 'min_window_size' is an integer, it must be greater than or equal "
      "to 1 and lower than or equal to n_timestamps (got 0)."),

     ({'min_window_size': -87}, ValueError,
      "If 'min_window_size' is an integer, it must be greater than or equal "
      "to 1 and lower than or equal to n_timestamps (got -87)."),

     ({'min_window_size': 32}, ValueError,
      "If 'min_window_size' is an integer, it must be greater than or equal "
      "to 1 and lower than or equal to n_timestamps (got 32)."),

     ({'min_window_size': 0.}, ValueError,
      "If 'min_window_size' is a float, it must be greater than 0 and "
      "lower than or equal to 1 (got 0.0)."),

     ({'min_window_size': 1.2}, ValueError,
      "If 'min_window_size' is a float, it must be greater than 0 and "
      "lower than or equal to 1 (got 1.2)."),

     ({'min_window_size': -0.3}, ValueError,
      "If 'min_window_size' is a float, it must be greater than 0 and "
      "lower than or equal to 1 (got -0.3).")]
)
def test_parameter_check(params, error, err_msg):
    """Test parameter validation."""
    fe = WindowFeatureExtractor(**params)
    with pytest.raises(error, match=re.escape(err_msg)):
        fe.fit(X_arange)


@pytest.mark.parametrize(
    'params',
    [{'min_window_size': 5},
     {'min_window_size': 15, 'n_windows': 80},
     {'min_window_size': 0.5},
     {'min_window_size': 0.5, 'n_windows': 20}]
)
def test_indices_window_feature_extractor(params):
    """Test that the indices are compatible with the input parameters."""
    fe = WindowFeatureExtractor(**params).fit(X_arange)
    indices = fe.indices_

    min_window_size = fe.get_params()['min_window_size']
    if isinstance(min_window_size, (float, np.floating)):
        min_window_size = np.ceil(min_window_size * n_timestamps)

    np.testing.assert_array_less(-1, indices[:, 0])
    np.testing.assert_array_less(indices[:, 1], n_timestamps + 1)
    np.testing.assert_array_less(min_window_size,
                                 indices[:, 1] - indices[:, 0] + 1)


@pytest.mark.parametrize(
    'params',
    [{'n_windows': 0.5},
     {'n_windows': 1.8},
     {'n_windows': 100, 'min_window_size': 0.2}]
)
def test_actual_results_random_indices(params):
    """Test the actual results with random indices."""
    fe = WindowFeatureExtractor(**params).fit(X_ones)

    n_windows = fe.get_params()['n_windows']
    if isinstance(n_windows, (float, np.floating)):
        n_windows = int(np.ceil(n_windows * n_timestamps))

    min_window_size = fe.get_params()['min_window_size']
    if isinstance(min_window_size, (float, np.floating)):
        min_window_size = int(np.ceil(min_window_size * n_timestamps))

    # With an array full of ones
    arr_actual = fe.transform(X_ones)
    arr_desired = np.tile(
        np.c_[np.ones(n_samples), np.zeros((n_samples, 2))], n_windows
    )
    np.testing.assert_allclose(arr_actual, arr_desired, atol=1e-5, rtol=0)

    # With an array arange-like, test the slope
    arr_actual = fe.transform(X_arange)
    if min_window_size > 1:
        np.testing.assert_allclose(arr_actual[:, 2::3], 1, atol=1e-5, rtol=0)
    else:
        assert np.all(np.isin(arr_actual[:, 2::3], [0, 1]))


@pytest.mark.parametrize(
    'indices, arr_desired',
    [([[0, 12]], [[2.125, 1.29301005, 0.25], [2.5, 1.32287566, 0]]),

     ([[0, 1], [0, 2], [8, 12]],
      [[0, 0, 0, 0.5, 0.5, 1, 3.25, 1.08972474, 0.9],
       [4, 0, 0, 3, 1, -2, 2.25, 1.08972474, 0.5]])]
)
def test_actual_results_fixed_indices(indices, arr_desired):
    """Test the actual results with fixed indices."""
    X = [[0, 1, 3, 2, 1, 3, 1.5, 1, 2, 3, 3, 5],
         [4, 2, 0, 2, 3, 5, 3.0, 2, 1, 2, 4, 2]]
    fe = WindowFeatureExtractor()
    fe.indices_ = np.asarray(indices)
    arr_actual = fe.transform(X)
    np.testing.assert_allclose(arr_actual, arr_desired, atol=1e-5, rtol=0)


@pytest.mark.parametrize('params', [{}, {'oob_score': True}, {'max_depth': 3}])
def test_attributes_time_series_forest(params):
    """Test the attributes of a fitted instance of TimeSeriesForest."""
    real_attributes = ['base_estimator_', 'classes_', 'estimators_',
                       'feature_importances_', 'indices_', 'n_features_in_',
                       'oob_decision_function_', 'oob_score_']
    fake_attributes = ['yolo', 'whoopsy', 'mistake_were_made_']

    clf = TimeSeriesForest(n_estimators=30, random_state=42, **params)
    clf.fit(X_arange, y)

    for attribute in real_attributes:
        assert hasattr(clf, attribute)

    for attribute in fake_attributes:
        assert not hasattr(clf, attribute)


def test_methods_time_series_forest():
    """Test the supported methods of TimeSeriesForest."""
    clf = TimeSeriesForest(n_estimators=7).fit(X_arange, y)
    assert clf.apply(X_arange).shape == (n_samples, 7)
    assert clf.decision_path(X_arange)[0].shape[0] == n_samples
    assert clf.decision_path(X_arange)[1].shape == (8,)
    assert clf.predict(X_arange).shape == (n_samples,)
    assert clf.predict_proba(X_arange).shape == (n_samples, 3)
    assert isinstance(clf.score(X_arange, y), (float, np.floating))



"""Testing for discretizers."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

import numpy as np
import pytest
import re
from pyts.preprocessing.discretizer import (
    _uniform_bins, _digitize, _reshape_with_nan
)
from pyts.preprocessing import KBinsDiscretizer


n_samples, n_timestamps = 3, 10
X = np.arange(n_samples * n_timestamps).reshape(n_samples, n_timestamps)


@pytest.mark.parametrize(
    'sample_min, sample_max, n_bins, arr_desired',
    [(np.arange(50), np.arange(20, 70), 2, [np.arange(10, 60)]),
     (np.arange(30, 80), np.arange(-10, 40), 2, [np.arange(10, 60)]),
     (np.arange(30), np.arange(30, 60), 3,
      [np.arange(10, 40), np.arange(20, 50)])]
)
def test_uniform_bins(sample_min, sample_max, n_bins, arr_desired):
    """Test '_uniform_bins' function."""
    arr_actual = _uniform_bins(sample_min, sample_max,
                               n_samples=sample_min.size, n_bins=n_bins)
    np.testing.assert_array_equal(arr_actual, arr_desired)


@pytest.mark.parametrize(
    'X, bins, arr_desired',
    [(X, np.array([4.5, 14.5, 24.5]),
      [[i] * 5 + [i + 1] * 5 for i in range(3)]),
     (X, np.array([[-0.5, 1.5, 8.5, 11.5],
                   [9.5, 11.5, 18.5, 21.5],
                   [19.5, 21.5, 28.5, 31.5]]),
      [[1] * 2 + [2] * 7 + [3] for _ in range(3)])]
)
def test_digitize(X, bins, arr_desired):
    """Test '_digitize' function."""
    X_float = np.asarray(X, dtype='float64')
    arr_actual = _digitize(X_float, bins)
    np.testing.assert_array_equal(arr_actual, arr_desired)


@pytest.mark.parametrize(
    'params, arr_desired',
    [({'X': (np.array([2., 4., 8.]), np.array([3., 5.])),
       'n_samples':2, 'lengths': np.array([3, 2]), 'max_length': 3},
      [[2, 4, 8], [3, 5, np.nan]]),

     ({'X': (np.array([2, 4, 8]), np.array([3]), np.array([5, 7])),
       'n_samples':3, 'lengths': np.array([3, 1, 2]), 'max_length': 3},
      [[2, 4, 8], [3, np.nan, np.nan], [5, 7, np.nan]]),

     ({'X': (np.array([3]), np.array([2, 4, 8]), np.array([5, 7])),
       'n_samples':3, 'lengths': np.array([1, 3, 2]), 'max_length': 3},
      [[3, np.nan, np.nan], [2, 4, 8], [5, 7, np.nan]])]
)
def test_reshape_with_nan(params, arr_desired):
    arr_actual = _reshape_with_nan(**params)
    np.testing.assert_array_equal(arr_actual, arr_desired)


@pytest.mark.parametrize(
    'params, error, err_msg',
    [({'n_bins': None}, TypeError, "'n_bins' must be an integer."),

     ({'n_bins': 1}, ValueError,
      "'n_bins' must be greater than or equal to 2 (got 1)."),

     ({'n_bins': 0}, ValueError,
      "'n_bins' must be greater than or equal to 2 (got 0)."),

     ({'strategy': 'whoops'}, ValueError,
      "'strategy' must be either 'uniform', 'quantile' or 'normal' (got {0})."
      .format('whoops'))]
)
def test_parameter_check(params, error, err_msg):
    """Test parameter validation."""
    discretizer = KBinsDiscretizer(**params)
    with pytest.raises(error, match=re.escape(err_msg)):
        discretizer.transform(X)


def test_warning_smaller_n_bins():
    """Test that a warning is raised when n_bins will be smaller."""
    X_new = np.r_[np.zeros((1, n_timestamps)), X]
    discretizer = KBinsDiscretizer()
    warning_msg = (
        "Some quantiles are equal. The number of bins will be "
        "smaller for sample {0}. Consider decreasing the number "
        "of bins or removing these samples.".format([0])
    )
    with pytest.warns(None, match=re.escape(warning_msg)):
        discretizer._compute_bins(
            X_new, n_samples + 1, n_bins=5, strategy='quantile'
        )


@pytest.mark.parametrize(
    'params, X, arr_desired',
    [({'n_bins': 10, 'strategy': 'quantile'}, X,
      [list(range(10))] * n_samples),

     ({'n_bins': 10, 'strategy': 'uniform'}, X,
      [list(range(10))] * n_samples),

     ({'n_bins': 5, 'strategy': 'quantile'}, X,
      [np.repeat(np.arange(5), 2)] * n_samples),

     ({'n_bins': 5, 'strategy': 'uniform'}, X,
      [np.repeat(np.arange(5), 2)] * n_samples),

     ({'n_bins': 3, 'strategy': 'normal'}, [[-0.5, 0, 0.5]], [[0, 1, 2]])]
)
def test_actual_results(params, X, arr_desired):
    """Test that the actual results are the expected ones."""
    arr_actual = KBinsDiscretizer(**params).fit_transform(X)
    np.testing.assert_allclose(arr_actual, arr_desired, atol=1e-5, rtol=0.)



"""Code for BOSS metric."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

import numpy as np
from math import sqrt
from sklearn.utils import check_array


def boss(x, y):
    r"""Return the BOSS distance between two arrays.

    Parameters
    ----------

    x : array-like, shape = (n_timestamps,)
        First array.

    y : array-like, shape = (n_timestamps,)
        Second array.

    Returns
    -------
    dist : float
        The BOSS distance between both arrays.

    Notes
    -----
    The BOSS metric is defined as

    .. math::

        BOSS(x, y) = \sum_{\substack{i=1\\ x_i > 0}}^n (x_i - y_i)^2

    where :math:`x` and :math:`y` are vectors of non-negative integers.
    The BOSS distance is not a distance metric as it neither satisfies the
    symmetry condition nor the triangle inequality.

    References
    ----------
    .. [1] P. Schäfer, "The BOSS is concerned with time series classification
           in the presence of noise". Data Mining and Knowledge Discovery,
           29(6), 1505-1530 (2015).

    Examples
    --------
    >>> from pyts.metrics import boss
    >>> x = [0, 5, 5, 3, 4, 5]
    >>> y = [3, 0, 0, 0, 8, 0]
    >>> boss(x, y)
    10.0
    >>> boss(y, x)
    5.0

    """
    x = check_array(x, ensure_2d=False, dtype='float64')
    y = check_array(y, ensure_2d=False, dtype='float64')
    if x.ndim != 1:
        raise ValueError("'x' must a one-dimensional array.")
    if y.ndim != 1:
        raise ValueError("'y' must a one-dimensional array.")
    if x.shape != y.shape:
        raise ValueError("'x' and 'y' must have the same shape.")

    non_zero_idx = ~np.isclose(x, np.zeros_like(x), rtol=1e-5, atol=1e-8)
    return sqrt(np.sum((x[non_zero_idx] - y[non_zero_idx]) ** 2))



"""The :mod:`pyts.utils` module includes utility tools."""

from .utils import segmentation, windowed_view
from .deprecation import deprecated


__all__ = ['segmentation', 'windowed_view', 'deprecated']



"""Testing for 'windowed_view' function."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

import numpy as np
import pytest
import re
from pyts.utils import windowed_view


X = np.arange(10).reshape(2, 5)


@pytest.mark.parametrize(
    'params, error, err_msg',
    [({'X': X, 'window_size': '3', 'window_step': 3}, TypeError,
      "'window_size' must be an integer."),
     ({'X': X, 'window_size': 3, 'window_step': '3'}, TypeError,
      "'window_step' must be an integer."),
     ({'X': X, 'window_size': 7, 'window_step': 3}, ValueError,
      "'window_size' must be an integer between 1 and n_timestamps."),
     ({'X': X, 'window_size': 3, 'window_step': 7}, ValueError,
      "'window_step' must be an integer between 1 and n_timestamps.")]
)
def test_parameter_check(params, error, err_msg):
    """Test parameter validation in segmentation."""
    with pytest.raises(error, match=re.escape(err_msg)):
        windowed_view(**params)


@pytest.mark.parametrize(
    'params, arr_desired',
    [({'X': X, 'window_size': 3, 'window_step': 1},
      [[[0, 1, 2], [1, 2, 3], [2, 3, 4]], [[5, 6, 7], [6, 7, 8], [7, 8, 9]]]),
     ({'X': X, 'window_size': 3, 'window_step': 2},
      [[[0, 1, 2], [2, 3, 4]], [[5, 6, 7], [7, 8, 9]]]),
     ({'X': X, 'window_size': 2, 'window_step': 3},
      [[[0, 1], [3, 4]], [[5, 6], [8, 9]]])]
)
def test_accurate_results(params, arr_desired):
    """Test that the actual results are the expected ones."""
    arr_actual = windowed_view(**params)
    np.testing.assert_array_equal(arr_actual, arr_desired)



"""Code for Word ExtrAction for time SEries cLassification."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

import numpy as np
from scipy.sparse import coo_matrix, csr_matrix, hstack
from sklearn.utils.validation import check_array, check_X_y, check_is_fitted
from sklearn.utils.multiclass import check_classification_targets
from sklearn.base import BaseEstimator
from sklearn.feature_extraction.text import CountVectorizer
from sklearn.feature_selection import chi2
from ..approximation import SymbolicFourierApproximation
from ..base import UnivariateTransformerMixin
from ..utils.utils import _windowed_view


class WEASEL(BaseEstimator, UnivariateTransformerMixin):
    """Word ExtrAction for time SEries cLassification.

    Parameters
    ----------
    word_size : int (default = 4)
        Size of each word.

    n_bins : int (default = 4)
        The number of bins to produce. It must be between 2 and 26.

    window_sizes : array-like (default = [0.1, 0.3, 0.5, 0.7, 0.9])
        Size of the sliding windows. All the elements must be either integers
        or floats. In the latter case, each element represents the percentage
        of the size of each time series and must be between 0 and 1; the size
        of the sliding windows will be computed as
        ``np.ceil(window_sizes * n_timestamps)``.

    window_steps : None or array-like (default = None)
        Step of the sliding windows. If None, each ``window_step`` is equal to
        ``window_size`` so that the windows are non-overlapping. Otherwise, all
        the elements must be either integers or floats. In the latter case,
        each element represents the percentage of the size of each time series
        and must be between 0 and 1; the step of the sliding windows will be
        computed as ``np.ceil(window_steps * n_timestamps)``.

    anova : bool (default = True)
        If True, the Fourier coefficient selection is done via a one-way
        ANOVA test. If False, the first Fourier coefficients are selected.

    drop_sum : bool (default = True)
        If True, the first Fourier coefficient (i.e. the sum of the subseries)
        is dropped. Otherwise, it is kept.

    norm_mean : bool (default = True)
        If True, center each subseries before scaling.

    norm_std : bool (default = True)
        If True, scale each subseries to unit variance.

    strategy : str (default = 'entropy')
        Strategy used to define the widths of the bins:

        - 'uniform': All bins in each sample have identical widths
        - 'quantile': All bins in each sample have the same number of points
        - 'normal': Bin edges are quantiles from a standard normal distribution
        - 'entropy': Bin edges are computed using information gain

    chi2_threshold : int or float (default = 2)
        The threshold used to perform feature selection. Only the words with
        a chi2 statistic above this threshold will be kept.

    sparse : bool (default = True)
        Return a sparse matrix if True, else return an array.

    alphabet : None, 'ordinal' or array-like, shape = (n_bins,)
        Alphabet to use. If None, the first `n_bins` letters of the Latin
        alphabet are used.

    Attributes
    ----------
    vocabulary_ : dict
        A mapping of features indices to terms.

    References
    ----------
    .. [1] P. Schäfer, and U. Leser, "Fast and Accurate Time Series
           Classification with WEASEL". Conference on Information and Knowledge
           Management, 637-646 (2017).

    Examples
    --------
    >>> from pyts.datasets import load_gunpoint
    >>> from pyts.transformation import WEASEL
    >>> X_train, _, y_train, _ = load_gunpoint(return_X_y=True)
    >>> weasel = WEASEL(sparse=False)
    >>> weasel.fit(X_train, y_train)
    WEASEL(...)
    >>> weasel.transform(X_train)
    array(...)

    """

    def __init__(self, word_size=4, n_bins=4,
                 window_sizes=[0.1, 0.3, 0.5, 0.7, 0.9], window_steps=None,
                 anova=True, drop_sum=True, norm_mean=True, norm_std=True,
                 strategy='entropy', chi2_threshold=2, sparse=True,
                 alphabet=None):
        self.word_size = word_size
        self.n_bins = n_bins
        self.window_sizes = window_sizes
        self.window_steps = window_steps
        self.anova = anova
        self.drop_sum = drop_sum
        self.norm_mean = norm_mean
        self.norm_std = norm_std
        self.strategy = strategy
        self.chi2_threshold = chi2_threshold
        self.sparse = sparse
        self.alphabet = alphabet

    def fit(self, X, y):
        """Fit the model according to the given training data.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_timestamps)
            Training vector.

        y : array-like, shape = (n_samples,)
            Class labels for each data sample.

        Returns
        -------
        self : object

        """
        X, y = check_X_y(X, y, dtype='float64')
        check_classification_targets(y)
        n_samples, n_timestamps = X.shape
        window_sizes, window_steps = self._check_params(n_timestamps)
        self._window_sizes = window_sizes
        self._window_steps = window_steps

        self._sfa_list = []
        self._vectorizer_list = []
        self._relevant_features_list = []
        self.vocabulary_ = {}

        for (window_size, window_step) in zip(window_sizes, window_steps):
            n_windows = ((n_timestamps - window_size + window_step)
                         // window_step)
            X_windowed = _windowed_view(
                X, n_samples, n_timestamps, window_size, window_step
            )
            X_windowed = X_windowed.reshape(n_samples * n_windows, window_size)

            sfa = SymbolicFourierApproximation(
                n_coefs=self.word_size, drop_sum=self.drop_sum,
                anova=self.anova, norm_mean=self.norm_mean,
                norm_std=self.norm_std, n_bins=self.n_bins,
                strategy=self.strategy, alphabet=self.alphabet
            )
            y_repeated = np.repeat(y, n_windows)
            X_sfa = sfa.fit_transform(X_windowed, y_repeated)

            X_word = np.asarray([''.join(X_sfa[i])
                                 for i in range(n_samples * n_windows)])
            X_word = X_word.reshape(n_samples, n_windows)

            X_bow = np.asarray([' '.join(X_word[i]) for i in range(n_samples)])
            vectorizer = CountVectorizer(ngram_range=(1, 2))
            X_counts = vectorizer.fit_transform(X_bow)
            chi2_statistics, _ = chi2(X_counts, y)
            relevant_features = np.where(
                chi2_statistics > self.chi2_threshold)[0]

            old_length_vocab = len(self.vocabulary_)
            vocabulary = {value: key
                          for (key, value) in vectorizer.vocabulary_.items()}
            for i, idx in enumerate(relevant_features):
                self.vocabulary_[i + old_length_vocab] = \
                    str(window_size) + " " + vocabulary[idx]

            self._relevant_features_list.append(relevant_features)
            self._sfa_list.append(sfa)
            self._vectorizer_list.append(vectorizer)

        return self

    def transform(self, X):
        """Transform the provided data.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_timestamps)
            Test samples.

        Returns
        -------
        X_new : sparse matrix, shape = (n_samples, n_features)
            Document-term matrix with relevant features only.

        """
        check_is_fitted(self, ['_relevant_features_list', '_sfa_list',
                               '_vectorizer_list', 'vocabulary_'])

        X = check_array(X, dtype='float64')
        n_samples, n_timestamps = X.shape

        X_features = coo_matrix((n_samples, 0), dtype=np.int64)

        for (window_size, window_step, sfa,
             vectorizer, relevant_features) in zip(
                 self._window_sizes, self._window_steps, self._sfa_list,
                 self._vectorizer_list, self._relevant_features_list):

            n_windows = ((n_timestamps - window_size + window_step)
                         // window_step)
            X_windowed = _windowed_view(
                X, n_samples, n_timestamps, window_size, window_step
            )
            X_windowed = X_windowed.reshape(n_samples * n_windows, window_size)
            X_sfa = sfa.transform(X_windowed)

            X_word = np.asarray([''.join(X_sfa[i])
                                 for i in range(n_samples * n_windows)])
            X_word = X_word.reshape(n_samples, n_windows)
            X_bow = np.asarray([' '.join(X_word[i]) for i in range(n_samples)])
            X_counts = vectorizer.transform(X_bow)[:, relevant_features]
            X_features = hstack([X_features, X_counts])

        if not self.sparse:
            return X_features.A
        return csr_matrix(X_features)

    def fit_transform(self, X, y):
        """Fit the data then transform it.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_timestamps)
            Train samples.

        y : array-like, shape = (n_samples,)
            Class labels for each data sample.

        Returns
        -------
        X_new : array, shape (n_samples, n_words)
            Document-term matrix.

        """
        X, y = check_X_y(X, y, dtype='float64')
        check_classification_targets(y)
        n_samples, n_timestamps = X.shape
        window_sizes, window_steps = self._check_params(n_timestamps)
        self._window_sizes = window_sizes
        self._window_steps = window_steps

        self._sfa_list = []
        self._vectorizer_list = []
        self._relevant_features_list = []
        self.vocabulary_ = {}

        X_features = coo_matrix((n_samples, 0), dtype=np.int64)

        for (window_size, window_step) in zip(window_sizes, window_steps):
            n_windows = ((n_timestamps - window_size + window_step)
                         // window_step)
            X_windowed = _windowed_view(
                X, n_samples, n_timestamps, window_size, window_step
            )
            X_windowed = X_windowed.reshape(n_samples * n_windows, window_size)

            sfa = SymbolicFourierApproximation(
                n_coefs=self.word_size, drop_sum=self.drop_sum,
                anova=self.anova, norm_mean=self.norm_mean,
                norm_std=self.norm_std, n_bins=self.n_bins,
                strategy=self.strategy, alphabet=self.alphabet
            )
            y_repeated = np.repeat(y, n_windows)
            X_sfa = sfa.fit_transform(X_windowed, y_repeated)

            X_word = np.asarray([''.join(X_sfa[i])
                                 for i in range(n_samples * n_windows)])
            X_word = X_word.reshape(n_samples, n_windows)

            X_bow = np.asarray([' '.join(X_word[i]) for i in range(n_samples)])
            vectorizer = CountVectorizer(ngram_range=(1, 2))
            X_counts = vectorizer.fit_transform(X_bow)
            chi2_statistics, _ = chi2(X_counts, y)
            relevant_features = np.where(
                chi2_statistics > self.chi2_threshold)[0]
            X_features = hstack([X_features, X_counts[:, relevant_features]])

            old_length_vocab = len(self.vocabulary_)
            vocabulary = {value: key
                          for (key, value) in vectorizer.vocabulary_.items()}
            for i, idx in enumerate(relevant_features):
                self.vocabulary_[i + old_length_vocab] = \
                    str(window_size) + " " + vocabulary[idx]

            self._relevant_features_list.append(relevant_features)
            self._sfa_list.append(sfa)
            self._vectorizer_list.append(vectorizer)

        if not self.sparse:
            return X_features.A
        return csr_matrix(X_features)

    def _check_params(self, n_timestamps):
        if not isinstance(self.word_size, (int, np.integer)):
            raise TypeError("'word_size' must be an integer.")
        if not self.word_size >= 1:
            raise ValueError("'word_size' must be a positive integer.")

        if not isinstance(self.window_sizes, (list, tuple, np.ndarray)):
            raise TypeError("'window_sizes' must be array-like.")
        window_sizes = check_array(self.window_sizes, ensure_2d=False,
                                   dtype=None)
        if window_sizes.ndim != 1:
            raise ValueError("'window_sizes' must be one-dimensional.")
        if not issubclass(window_sizes.dtype.type, (np.integer, np.floating)):
            raise ValueError("The elements of 'window_sizes' must be integers "
                             "or floats.")
        if issubclass(window_sizes.dtype.type, np.floating):
            if not (np.min(window_sizes) > 0 and np.max(window_sizes) <= 1):
                raise ValueError(
                    "If the elements of 'window_sizes' are floats, they all "
                    "must be greater than 0 and lower than or equal to 1."
                )
            window_sizes = np.ceil(window_sizes * n_timestamps).astype('int64')
        if not np.max(window_sizes) <= n_timestamps:
            raise ValueError("All the elements in 'window_sizes' must be "
                             "lower than or equal to n_timestamps.")

        if self.drop_sum and not self.word_size < np.min(window_sizes):
            raise ValueError(
                "If 'drop_sum=True', 'word_size' must be lower than "
                "the minimum value in 'window_sizes'."
            )
        if not (self.drop_sum or self.word_size <= np.min(window_sizes)):
            raise ValueError(
                "If 'drop_sum=False', 'word_size' must be lower than or "
                "equal to the minimum value in 'window_sizes'."
            )

        if not ((self.window_steps is None)
                or isinstance(self.window_steps, (list, tuple, np.ndarray))):
            raise TypeError("'window_steps' must be None or array-like.")
        if self.window_steps is None:
            window_steps = window_sizes
        else:
            window_steps = check_array(self.window_steps, ensure_2d=False,
                                       dtype=None)
            if window_steps.ndim != 1:
                raise ValueError("'window_steps' must be one-dimensional.")
            if window_steps.size != window_sizes.size:
                raise ValueError("If 'window_steps' is not None, it must have "
                                 "the same size as 'window_sizes'.")
            if not issubclass(window_steps.dtype.type,
                              (np.integer, np.floating)):
                raise ValueError(
                    "If 'window_steps' is not None, the elements of "
                    "'window_steps' must be integers or floats."
                )
            if issubclass(window_steps.dtype.type, np.floating):
                if not ((np.min(window_steps) > 0
                         and np.max(window_steps) <= 1)):
                    raise ValueError(
                        "If the elements of 'window_steps' are floats, they "
                        "all must be greater than 0 and lower than or equal "
                        "to 1."
                    )
                window_steps = np.ceil(
                    window_steps * n_timestamps).astype('int64')
            if not ((np.min(window_steps) >= 1)
                    and (np.max(window_steps) <= n_timestamps)):
                raise ValueError("All the elements in 'window_steps' must be "
                                 "greater than or equal to 1 and lower than "
                                 "or equal to n_timestamps.")

        if not isinstance(self.chi2_threshold,
                          (int, np.integer, float, np.floating)):
            raise TypeError("'chi2_threshold' must be a float or an "
                            "integer.")
        if not self.chi2_threshold > 0:
            raise ValueError("'chi2_threshold' must be positive.")
        return window_sizes, window_steps



"""Testing for Symbolic Fourier Approximation."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

import numpy as np
import pytest
from sklearn.feature_selection import f_classif
from pyts.approximation import MultipleCoefficientBinning
from pyts.approximation import SymbolicFourierApproximation


rng = np.random.RandomState(42)
n_samples, n_timestamps = 5, 8
X = rng.randn(n_samples, n_timestamps)
y = rng.randint(2, size=n_samples)


def _compute_expected_results(X, y=None, n_coefs=None, n_bins=4,
                              strategy='quantile', drop_sum=False, anova=False,
                              norm_mean=False, norm_std=False, alphabet=None):
    """Compute the expected results."""
    X = np.asarray(X)
    if norm_mean:
        X -= X.mean(axis=1)[:, None]
    if norm_std:
        X /= X.std(axis=1)[:, None]
    X_fft = np.fft.rfft(X)
    X_fft = np.vstack([np.real(X_fft), np.imag(X_fft)])
    X_fft = X_fft.reshape(n_samples, -1, order='F')
    if drop_sum:
        X_fft = X_fft[:, 2:-1]
    else:
        X_fft = np.hstack([X_fft[:, :1], X_fft[:, 2:-1]])
    if n_coefs is not None:
        if anova:
            _, p = f_classif(X_fft, y)
            support = np.argsort(p)[:n_coefs]
            X_fft = X_fft[:, support]
        else:
            X_fft = X_fft[:, :n_coefs]

    mcb = MultipleCoefficientBinning(n_bins=n_bins, strategy=strategy,
                                     alphabet=alphabet)
    arr_desired = mcb.fit_transform(X_fft)
    return arr_desired


@pytest.mark.parametrize(
    'params',
    [({}),
     ({'n_coefs': 3}),
     ({'n_bins': 2}),
     ({'strategy': 'uniform'}),
     ({'drop_sum': True}),
     ({'anova': True}),
     ({'norm_mean': True, 'drop_sum': True}),
     ({'norm_std': True}),
     ({'norm_mean': True, 'norm_std': True, 'drop_sum': True}),
     ({'n_coefs': 2, 'drop_sum': True, 'anova': True})]
)
def test_actual_results(params):
    """Test that the actual results are the expected ones."""
    arr_actual = SymbolicFourierApproximation(**params).fit_transform(X, y)
    arr_desired = _compute_expected_results(X, y, **params)
    np.testing.assert_array_equal(arr_actual, arr_desired)


@pytest.mark.parametrize(
    'params',
    [({}),
     ({'n_coefs': 3}),
     ({'n_bins': 2}),
     ({'strategy': 'uniform'}),
     ({'drop_sum': True}),
     ({'anova': True}),
     ({'norm_mean': True, 'drop_sum': True}),
     ({'norm_std': True}),
     ({'norm_mean': True, 'norm_std': True, 'drop_sum': True}),
     ({'n_coefs': 2, 'drop_sum': True, 'anova': True})]
)
def test_fit_transform(params):
    """Test that fit and transform yield the same results as fit_transform."""
    arr_1 = SymbolicFourierApproximation(**params).fit(X, y).transform(X)
    arr_2 = SymbolicFourierApproximation(**params).fit_transform(X, y)
    np.testing.assert_array_equal(arr_1, arr_2)



"""Code for Multiple Coefficient Binning."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

import numpy as np
from numba import njit, prange
from scipy.stats import norm
from sklearn.base import BaseEstimator
from ..base import UnivariateTransformerMixin
from sklearn.tree import DecisionTreeClassifier
from sklearn.utils.validation import check_array, check_is_fitted, check_X_y
from sklearn.utils.multiclass import check_classification_targets


@njit()
def _uniform_bins(timestamp_min, timestamp_max, n_timestamps, n_bins):
    bin_edges = np.empty((n_timestamps, n_bins - 1))
    for i in prange(n_timestamps):
        bin_edges[i] = np.linspace(
            timestamp_min[i], timestamp_max[i], n_bins + 1
        )[1:-1]
    return bin_edges


@njit()
def _digitize_1d(X, bins, n_samples, n_timestamps):
    X_digit = np.empty((n_samples, n_timestamps))
    for i in prange(n_timestamps):
        X_digit[:, i] = np.digitize(X[:, i], bins, right=True)
    return X_digit


@njit()
def _digitize_2d(X, bins, n_samples, n_timestamps):
    X_digit = np.empty((n_samples, n_timestamps))
    for i in prange(n_timestamps):
        X_digit[:, i] = np.digitize(X[:, i], bins[i], right=True)
    return X_digit


def _digitize(X, bins):
    n_samples, n_timestamps = X.shape
    if bins.ndim == 1:
        X_binned = _digitize_1d(X, bins, n_samples, n_timestamps)
    else:
        X_binned = _digitize_2d(X, bins, n_samples, n_timestamps)
    return X_binned.astype('int64')


class MultipleCoefficientBinning(BaseEstimator, UnivariateTransformerMixin):
    """Bin continuous data into intervals column-wise.

    Parameters
    ----------
    n_bins : int (default = 4)
        The number of bins to produce. It must be between 2 and
        ``min(n_samples, 26)``.

    strategy : str (default = 'quantile')
        Strategy used to define the widths of the bins:

        - 'uniform': All bins in each sample have identical widths
        - 'quantile': All bins in each sample have the same number of points
        - 'normal': Bin edges are quantiles from a standard normal distribution
        - 'entropy': Bin edges are computed using information gain

    alphabet : None, 'ordinal' or array-like, shape = (n_bins,)
        Alphabet to use. If None, the first `n_bins` letters of the Latin
        alphabet are used if `n_bins` is lower than 27, otherwise the alphabet
        will be defined to [chr(i) for i in range(n_bins)]. If 'ordinal',
        integers are used.

    Attributes
    ----------
    bin_edges_ : array, shape = (n_bins - 1,) or (n_timestamps, n_bins - 1)
        Bin edges with shape = (n_bins - 1,) if ``strategy='normal'`` or
        (n_timestamps, n_bins - 1) otherwise.

    References
    ----------
    .. [1] P. Schäfer, and M. Högqvist, "SFA: A Symbolic Fourier Approximation
           and Index for Similarity Search in High Dimensional Datasets",
           International Conference on Extending Database Technology,
           15, 516-527 (2012).

    Examples
    --------
    >>> from pyts.approximation import MultipleCoefficientBinning
    >>> X = [[0, 4],
    ...      [2, 7],
    ...      [1, 6],
    ...      [3, 5]]
    >>> transformer = MultipleCoefficientBinning(n_bins=2)
    >>> print(transformer.fit_transform(X))
    [['a' 'a']
     ['b' 'b']
     ['a' 'b']
     ['b' 'a']]

    """

    def __init__(self, n_bins=4, strategy='quantile', alphabet=None):
        self.n_bins = n_bins
        self.strategy = strategy
        self.alphabet = alphabet

    def fit(self, X, y=None):
        """Compute the bin edges for each feature.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_timestamps)
            Data to transform.

        y : None or array-like, shape = (n_samples,)
            Class labels for each sample. Only used if ``strategy='entropy'``.

        """
        if self.strategy == 'entropy':
            if y is None:
                raise ValueError("y cannot be None if strategy='entropy'.")
            X, y = check_X_y(X, y, dtype='float64')
            check_classification_targets(y)
        else:
            X = check_array(X, dtype='float64')
        n_samples, n_timestamps = X.shape
        self._n_timestamps_fit = n_timestamps
        self._alphabet = self._check_params(n_samples)
        self._check_constant(X)
        self.bin_edges_ = self._compute_bins(
            X, y, n_timestamps, self.n_bins, self.strategy)
        return self

    def transform(self, X):
        """Bin the data.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_timestamps)
            Data to transform.

        Returns
        -------
        X_new : array, shape = (n_samples, n_timestamps)
            Binned data.

        """
        check_is_fitted(self, 'bin_edges_')
        X = check_array(X, dtype='float64')
        self._check_consistent_lengths(X)
        indices = _digitize(X, self.bin_edges_)
        if isinstance(self._alphabet, str):
            return indices
        else:
            return self._alphabet[indices]

    def _check_params(self, n_samples):
        if not isinstance(self.n_bins, (int, np.integer)):
            raise TypeError("'n_bins' must be an integer.")
        if not 2 <= self.n_bins <= min(n_samples, 26):
            raise ValueError(
                "'n_bins' must be greater than or equal to 2 and lower than "
                "or equal to min(n_samples, 26) (got {0}).".format(self.n_bins)
            )
        if self.strategy not in ['uniform', 'quantile', 'normal', 'entropy']:
            raise ValueError("'strategy' must be either 'uniform', 'quantile',"
                             " 'normal' or 'entropy' (got {0})."
                             .format(self.strategy))
        if not ((self.alphabet is None)
                or (self.alphabet == 'ordinal')
                or (isinstance(self.alphabet, (list, tuple, np.ndarray)))):
            raise TypeError("'alphabet' must be None, 'ordinal' or array-like "
                            "with shape (n_bins,) (got {0})."
                            .format(self.alphabet))
        if self.alphabet is None:
            alphabet = np.array([chr(i) for i in range(97, 97 + self.n_bins)])
        elif self.alphabet == 'ordinal':
            alphabet = 'ordinal'
        else:
            alphabet = check_array(self.alphabet, ensure_2d=False, dtype=None)
            if alphabet.shape != (self.n_bins,):
                raise ValueError("If 'alphabet' is array-like, its shape "
                                 "must be equal to (n_bins,).")
        return alphabet

    def _check_constant(self, X):
        if np.any(np.max(X, axis=0) - np.min(X, axis=0) == 0):
            raise ValueError("At least one timestamp is constant.")

    def _check_consistent_lengths(self, X):
        if self._n_timestamps_fit != X.shape[1]:
            raise ValueError(
                "The number of timestamps in X must be the same as "
                "the number of timestamps when `fit` was called "
                "({0} != {1}).".format(self._n_timestamps_fit, X.shape[1])
            )

    def _compute_bins(self, X, y, n_timestamps, n_bins, strategy):
        if strategy == 'normal':
            bins_edges = norm.ppf(np.linspace(0, 1, self.n_bins + 1)[1:-1])
        elif strategy == 'uniform':
            timestamp_min, timestamp_max = np.min(X, axis=0), np.max(X, axis=0)
            bins_edges = _uniform_bins(timestamp_min, timestamp_max,
                                       n_timestamps, n_bins)
        elif strategy == 'quantile':
            bins_edges = np.percentile(
                X, np.linspace(0, 100, self.n_bins + 1)[1:-1], axis=0
            ).T
            if np.any(np.diff(bins_edges, axis=0) == 0):
                raise ValueError(
                    "At least two consecutive quantiles are equal. "
                    "Consider trying with a smaller number of bins or "
                    "removing timestamps with low variation."
                )
        else:
            bins_edges = self._entropy_bins(X, y, n_timestamps, n_bins)
        return bins_edges

    def _entropy_bins(self, X, y, n_timestamps, n_bins):
        bins = np.empty((n_timestamps, n_bins - 1))
        clf = DecisionTreeClassifier(criterion='entropy',
                                     max_leaf_nodes=n_bins)
        for i in range(n_timestamps):
            clf.fit(X[:, i][:, None], y)
            threshold = clf.tree_.threshold[clf.tree_.children_left != -1]
            if threshold.size < (n_bins - 1):
                raise ValueError(
                    "The number of bins is too high for timestamp {0}. "
                    "Consider trying with a smaller number of bins or "
                    "removing this timestamp.".format(i)
                )
            bins[i] = threshold
        return np.sort(bins, axis=1)



"""Time series transformation and classification module for Python.

pyts is a Python package for time series classification.
It aims to make time series classification easily accessible by providing
preprocessing and utility tools, and implementations of state-of-the-art
algorithms. Most of these algorithms transform time series, thus pyts provides
several tools to perform these transformations.
"""

__version__ = '0.12.0'

__all__ = ['approximation', 'bag_of_words', 'classification', 'datasets',
           'decomposition', 'image', 'metrics', 'multivariate',
           'preprocessing', 'transformation', 'utils']



"""Testing for JointRecurrencePlot."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

import numpy as np
import pytest
import re
from pyts.image import RecurrencePlot
from pyts.multivariate.image import JointRecurrencePlot


n_samples, n_features, n_timestamps = 40, 3, 30
rng = np.random.RandomState(42)
X = rng.randn(n_samples, n_features, n_timestamps)


@pytest.mark.parametrize(
    'params, error, err_msg',
    [({'threshold': [None, None]}, ValueError,
      "If 'threshold' is a list, its length must be equal to n_features "
      "(2 != 3)."),

     ({'percentage': [0, 2, 4, 5]}, ValueError,
      "If 'percentage' is a list, its length must be equal to n_features "
      "(4 != 3).")]
)
def test_parameter_check(params, error, err_msg):
    """Test parameter validation."""
    transformer = JointRecurrencePlot(**params)
    with pytest.raises(error, match=re.escape(err_msg)):
        transformer.transform(X)


@pytest.mark.parametrize(
    'params, shape_desired',
    [({}, (40, 30, 30)),
     ({'dimension': 3}, (40, 28, 28)),
     ({'dimension': 9}, (40, 22, 22)),
     ({'time_delay': 3}, (40, 30, 30)),
     ({'dimension': 3, 'time_delay': 4}, (40, 22, 22)),
     ({'dimension': 6, 'time_delay': 3}, (40, 15, 15))]
)
def test_shapes(params, shape_desired):
    """Test that the shape of the output is the expected one."""
    transformer = JointRecurrencePlot(**params)
    assert transformer.fit(X).transform(X).shape == shape_desired
    assert transformer.fit_transform(X).shape == shape_desired


@pytest.mark.parametrize(
    'params',
    [{},
     {'dimension': 3},
     {'time_delay': 3},
     {'dimension': 3, 'time_delay': 4},
     {'dimension': 6, 'time_delay': 3},
     {'threshold': 3},
     {'percentage': 30}]
)
def test_actual_results_single_value(params):
    """Test that the actual results are the expected ones."""
    arr_actual = JointRecurrencePlot(**params).transform(X)
    arr_desired = []
    for i in range(n_features):
        arr_desired.append(RecurrencePlot(**params).transform(X[:, i]))
    arr_desired = np.prod(arr_desired, axis=0)
    np.testing.assert_allclose(arr_actual, arr_desired, atol=1e-5, rtol=0.)


@pytest.mark.parametrize(
    'params',
    [{'threshold': [0.5, 0.9, 2], 'percentage': [10, 30, 50]},
     {'threshold': [None, 'distance', 'point'], 'percentage': [10, 30, 50]},
     {'threshold': [0.5, 0.9, None], 'percentage': [10., 30., 90.]}]
)
def test_actual_results_lists(params):
    """Test that the actual results are the expected ones."""
    arr_actual = JointRecurrencePlot(**params).transform(X)
    arr_desired = []
    for i, (threshold, percentage) in enumerate(
        zip(params['threshold'], params['percentage'])
    ):
        arr_desired.append(RecurrencePlot(
            threshold=threshold, percentage=percentage).transform(X[:, i]))
    arr_desired = np.prod(arr_desired, axis=0)
    np.testing.assert_allclose(arr_actual, arr_desired, atol=1e-5, rtol=0.)



"""Code for Dynamic Time Warping and its variants."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

import numpy as np
from math import ceil, log2, sqrt
from numba import njit, prange
from sklearn.utils import check_array


@njit()
def _square(x, y):
    return (x - y) ** 2


@njit()
def _absolute(x, y):
    return abs(x - y)


@njit()
def _cost_matrix_region(x, y, dist, region):
    n_timestamps_1, n_timestamps_2 = x.size, y.size
    cost_mat = np.full((n_timestamps_1, n_timestamps_2), np.inf)
    for i in prange(n_timestamps_1):
        for j in prange(region[0, i], region[1, i]):
            cost_mat[i, j] = dist(x[i], y[j])
    return cost_mat


@njit()
def _project_cost_matrix_region(cost_mat, region):
    n_timestamps_1, n_timestamps_2 = cost_mat.shape
    cost_mat_projected = np.full((n_timestamps_1, n_timestamps_2), np.inf)
    for i in prange(n_timestamps_1):
        for j in prange(region[0, i], region[1, i]):
            cost_mat_projected[i, j] = cost_mat[i, j]
    return cost_mat_projected


@njit()
def _cost_matrix_no_region(x, y, dist):
    n_timestamps_1, n_timestamps_2 = x.size, y.size
    cost_mat = np.empty((n_timestamps_1, n_timestamps_2))
    for j in prange(n_timestamps_2):
        for i in prange(n_timestamps_1):
            cost_mat[i, j] = dist(x[i], y[j])
    return cost_mat


def _check_input_dtw(x, y, precomputed_cost, dist, method):
    if dist == "precomputed" and method in ["multiscale", "fast"]:
        raise ValueError("The method '{0}' cannot be used with "
                         "a precomputed cost. Provide the raw time series or"
                         "use one of the methods: 'classic', 'sakoechiba', "
                         "'itakura'".format(method))
    if dist == "precomputed":
        cost = check_array(precomputed_cost, ensure_min_samples=2,
                           ensure_min_features=2, ensure_2d=True,
                           force_all_finite=False, dtype='float64')
        n_timestamps_1, n_timestamps_2 = precomputed_cost.shape
    else:
        cost = None
        x = check_array(x, ensure_2d=False, dtype='float64')
        y = check_array(y, ensure_2d=False, dtype='float64')
        if x.ndim != 1:
            raise ValueError("'x' must be a one-dimensional array.")
        if y.ndim != 1:
            raise ValueError("'y' must be a one-dimensional array.")
        n_timestamps_1 = x.size
        n_timestamps_2 = y.size

    return x, y, cost, n_timestamps_1, n_timestamps_2


def _input_to_cost(x, y, dist, precomputed_cost, region):
    """Computes cost matrix from dtw input."""
    if dist == "precomputed":
        if region is not None:
            cost_mat = _project_cost_matrix_region(precomputed_cost, region)
        else:
            cost_mat = precomputed_cost.copy()
    else:
        cost_mat = cost_matrix(x, y, dist=dist, region=region)
    cost_mat = check_array(cost_mat, ensure_min_samples=2,
                           ensure_min_features=2, ensure_2d=True,
                           force_all_finite=False, dtype='float64')
    return cost_mat


def _check_region(region, n_timestamps_1, n_timestamps_2):
    """Project region on the feasible set."""
    region = np.clip(region[:, :n_timestamps_1], 0, n_timestamps_2)
    return region


def cost_matrix(x, y, dist='square', region=None):
    """Compute the cost matrix between two samples.

    Parameters
    ----------
    x : array-like, shape = (n_timestamps_1,)
        First sample.

    y : array-like, shape = (n_timestamps_2,)
        Second sample.

    dist : 'square', 'absolute' or callable (default = 'square')
        Distance used. If 'square', the squared difference is used.
        If 'absolute', the absolute difference is used. If callable,
        it must be a function with a numba.njit() decorator that takes
        as input two numbers (two arguments) and returns a number.

    region : None or array-like, shape = (2, n_timestamps_1) (default = None)
        Constraint region. If None, there is no contraint region.
        If array-like, the first row indicates the starting indices (included)
        and the second row the ending indices (excluded) of the valid rows
        for each column.

    Returns
    -------
    cost_matrix : array, shape = (n_timestamps_1, n_timestamps_2)
        Cost matrix.

    """
    if dist == 'square':
        dist_ = _square
    elif dist == 'absolute':
        dist_ = _absolute
    elif isinstance(dist, str):
        raise ValueError("'dist' must be either 'square', 'absolute' or "
                         "callable (got {0}).".format(dist))
    else:
        try:
            temp = dist(1, 1)
        except:  # noqa: E722
            raise ValueError("Calling dist(1, 1) did not work.")
        else:
            if not isinstance(temp, (int, float)):
                raise ValueError("Calling dist(1, 1) did not return a float "
                                 "or an integer.")
        dist_ = dist
    if region is not None:
        region = check_array(region)
        region_shape = region.shape
        if region_shape != (2, x.size):
            raise ValueError(
                "The shape of 'region' must be equal to (2, n_timestamps_1) "
                "(got {0}).".format(region_shape)
            )
    if region is None:
        cost_mat = _cost_matrix_no_region(x, y, dist_)
    else:
        cost_mat = _cost_matrix_region(x, y, dist_, region)
    return cost_mat


@njit()
def _accumulated_cost_matrix_region(cost_matrix, region):
    n_timestamps_1, n_timestamps_2 = cost_matrix.shape
    acc_cost_mat = np.ones((n_timestamps_1, n_timestamps_2)) * np.inf
    acc_cost_mat[0, 0: region[1, 0]] = np.cumsum(
        cost_matrix[0, 0: region[1, 0]]
    )
    acc_cost_mat[0: region[1, 0], 0] = np.cumsum(
        cost_matrix[0: region[1, 0], 0]
    )
    region_ = np.copy(region)
    region_[0] = np.maximum(region_[0], 1)
    for i in range(1, n_timestamps_1):
        for j in range(region_[0, i], region_[1, i]):
            acc_cost_mat[i, j] = cost_matrix[i, j] + min(
                acc_cost_mat[i - 1][j - 1],
                acc_cost_mat[i - 1][j],
                acc_cost_mat[i][j - 1]
            )
    return acc_cost_mat


@njit()
def _accumulated_cost_matrix_no_region(cost_matrix):
    n_timestamps_1, n_timestamps_2 = cost_matrix.shape
    acc_cost_mat = np.empty((n_timestamps_1, n_timestamps_2))
    acc_cost_mat[0] = np.cumsum(cost_matrix[0])
    acc_cost_mat[:, 0] = np.cumsum(cost_matrix[:, 0])
    for j in range(1, n_timestamps_2):
        for i in range(1, n_timestamps_1):
            acc_cost_mat[i, j] = cost_matrix[i, j] + min(
                acc_cost_mat[i - 1][j - 1],
                acc_cost_mat[i - 1][j],
                acc_cost_mat[i][j - 1]
            )
    return acc_cost_mat


def accumulated_cost_matrix(cost_mat, region=None):
    """Compute the accumulated cost matrix.

    Parameters
    ----------
    cost_mat : array-like, shape = (n_timestamps_1, n_timestamps_2)
        Cost matrix.

    region : None or tuple, shape = (2, n_timestamps_1) (default = None)
        Constraint region. If None, there is no contraint region.
        If array-like, the first row indicates the starting indices (included)
        and the second row the ending indices (excluded) of the valid rows
        for each column.

    Returns
    -------
    acc_cost_mat : array, shape = (n_timestamps_1, n_timestamps_2)
        Accumulated cost matrix.

    """
    cost_mat = check_array(cost_mat, ensure_min_samples=2,
                           ensure_min_features=2, ensure_2d=True,
                           force_all_finite=False, dtype='float64')
    cost_mat_shape = cost_mat.shape

    if region is None:
        acc_cost_mat = _accumulated_cost_matrix_no_region(cost_mat)
    else:
        region = check_array(region, dtype='int64')
        region_shape = region.shape
        if region_shape != (2, cost_mat_shape[0]):
            raise ValueError("The shape of 'region' must be equal to "
                             "(2, n_timestamps_1) "
                             "(got {0})".format(region_shape)
                             )
        acc_cost_mat = _accumulated_cost_matrix_region(cost_mat, region)
    return acc_cost_mat


@njit()
def _return_path(acc_cost_mat):
    n_timestamps_1, n_timestamps_2 = acc_cost_mat.shape
    path = [(n_timestamps_1 - 1, n_timestamps_2 - 1)]
    while path[-1] != (0, 0):
        i, j = path[-1]
        if i == 0:
            path.append((0, j - 1))
        elif j == 0:
            path.append((i - 1, 0))
        else:
            arr = np.array([acc_cost_mat[i - 1][j - 1],
                            acc_cost_mat[i - 1][j],
                            acc_cost_mat[i][j - 1]])
            argmin = np.argmin(arr)
            if argmin == 0:
                path.append((i - 1, j - 1))
            elif argmin == 1:
                path.append((i - 1, j))
            else:
                path.append((i, j - 1))
    return np.transpose(np.array(path)[::-1])


def _return_results(dtw_dist, cost_mat, acc_cost_mat,
                    return_cost=False, return_accumulated=False,
                    return_path=False):
    """Return the results according to the parameters."""
    res = (dtw_dist, )
    if return_cost:
        res += (cost_mat, )
    if return_accumulated:
        res += (acc_cost_mat, )
    if return_path:
        path = _return_path(acc_cost_mat)
        res += (path, )
    if len(res) == 1:
        return res[0]
    else:
        return res


def _dtw_classic(x=None, y=None, dist='square', precomputed_cost=None,
                 return_cost=False, return_accumulated=False,
                 return_path=False):
    """Classic Dynamic Time Warping (DTW) distance between two time series.

    The classic version of DTW has no constraint region, allowing for possibly
    very large time shifts. This can be an upside or a downside depending on
    the application. If you do not want to allow large time shifts,
    consider using another method with a constraint region.

    Options
    -------
    This method has no options.

    References
    ----------
    .. [1] H. Sakoe and S. Chiba, "Dynamic programming algorithm optimization
           for spoken word recognition". IEEE Transactions on Acoustics,
           Speech, and Signal Processing, 26(1), 43-49 (1978).

    """
    x, y, precomputed_cost, n_timestamps_1, n_timestamps_2 = \
        _check_input_dtw(x, y, precomputed_cost, dist, method="classic")
    cost_mat = _input_to_cost(x, y, dist, precomputed_cost, region=None)
    acc_cost_mat = accumulated_cost_matrix(cost_mat, region=None)
    dtw_dist = acc_cost_mat[-1, -1]
    if dist == 'square':
        dtw_dist = sqrt(dtw_dist)

    res = _return_results(dtw_dist, cost_mat, acc_cost_mat,
                          return_cost, return_accumulated, return_path)
    return res


def _dtw_region(x=None, y=None, dist='square', region=None,
                precomputed_cost=None, return_cost=False,
                return_accumulated=False, return_path=False):
    """Dynamic Time Warping (DTW) distance with a constraint region.

    This version of DTW allows users to provide their own constraint region.
    If not provided, no constraint region is used, which is exactly the
    classic DTW.

    Options
    -------
    region : None or array-like, shape = (2, n_timestamps_1) (default = None)
        Constraint region. If None, no constraint region is used. Otherwise,
        the first row consists of the starting indices (included) and the
        second row consists of the ending indices (excluded) of the valid rows
        for each column.

    """
    x, y, precomputed_cost, n_timestamps_1, n_timestamps_2 = \
        _check_input_dtw(x, y, precomputed_cost, dist, method="region")
    cost_mat = _input_to_cost(x, y, dist, precomputed_cost, region=region)
    acc_cost_mat = accumulated_cost_matrix(cost_mat, region)
    dtw_dist = acc_cost_mat[-1, -1]
    if dist == 'square':
        dtw_dist = sqrt(dtw_dist)

    res = _return_results(dtw_dist, cost_mat, acc_cost_mat,
                          return_cost, return_accumulated, return_path)
    return res


def _check_sakoe_chiba_params(n_timestamps_1, n_timestamps_2, window_size):
    """Check and set some parameters of the sakoe-chiba band."""
    if not isinstance(n_timestamps_1, (int, np.integer)):
        raise TypeError("'n_timestamps_1' must be an integer.")
    else:
        if not n_timestamps_1 >= 2:
            raise ValueError("'n_timestamps_1' must be an integer greater than"
                             " or equal to 2.")
    if not isinstance(window_size, (int, np.integer, float, np.floating)):
        raise TypeError("'window_size' must be an integer or a float.")
    n_timestamps = max(n_timestamps_1, n_timestamps_2)

    if isinstance(window_size, (float, np.floating)):
        if not 0. <= window_size <= 1.:
            raise ValueError("The given 'window_size' is a float, "
                             "it must be between "
                             "0. and 1. To set the size of the sakoe-chiba "
                             "manually, 'window_size' must be an integer.")
        window_size = ceil(window_size * (n_timestamps - 1))
    else:
        if not 0 <= window_size <= (n_timestamps - 1):
            raise ValueError(
                "The given 'window_size' is an integer, it must "
                "be greater "
                "than or equal to 0 and lower than max('n_timestamps_1', "
                "'n_timestamps_2')."
            )

    scale = (n_timestamps_2 - 1) / (n_timestamps_1 - 1)

    if n_timestamps_2 > n_timestamps_1:
        window_size = max(window_size, scale / 2)
        horizontal_shift = 0
        vertical_shift = window_size
    elif n_timestamps_1 > n_timestamps_2:
        window_size = max(window_size, 0.5 / scale)
        horizontal_shift = window_size
        vertical_shift = 0
    else:
        horizontal_shift = 0
        vertical_shift = window_size
    return scale, horizontal_shift, vertical_shift


def sakoe_chiba_band(n_timestamps_1, n_timestamps_2=None, window_size=0.1):
    """Compute the Sakoe-Chiba band.

    Parameters
    ----------
    n_timestamps_1 : int
        The size of the first time series.

    n_timestamps_2 : int (optional, default None)
        The size of the second time series. If None, set to `n_timestamps_1`.

    window_size : float or int (default = 0.1)
        The window size above and below the diagonale.
        If float, `window_size must be between 0 and
        1, and the actual window size will be computed as:
        ``ceil(window_size * max((n_timestamps_1, n_timestamps_2) - 1))``.
        If int, `window_size` must be the largest temporal shift allowed.
        Each cell whose distance with the diagonale is lower than or equal to
        'window_size' becomes a valid cell for the path.

    Returns
    -------
    region : array, shape = (2, n_timestamps_1)
        Constraint region. The first row consists of the starting indices
        (included) and the second row consists of the ending indices (excluded)
        of the valid rows for each column.

    References
    ----------
    .. [1] H. Sakoe and S. Chiba, “Dynamic programming algorithm optimization
           for spoken word recognition”. IEEE Transactions on Acoustics,
           Speech, and Signal Processing, 26(1), 43-49 (1978).

    Examples
    --------
    >>> from pyts.metrics import sakoe_chiba_band
    >>> print(sakoe_chiba_band(5, window_size=0.5))
    [[0 0 0 1 2]
     [3 4 5 5 5]]

    """
    if n_timestamps_2 is None:
        n_timestamps_2 = n_timestamps_1
    scale, horizontal_shift, vertical_shift = \
        _check_sakoe_chiba_params(n_timestamps_1, n_timestamps_2, window_size)

    lower_bound = scale * (np.arange(n_timestamps_1) - horizontal_shift) \
        - vertical_shift
    lower_bound = np.round(lower_bound, 2)
    lower_bound = np.ceil(lower_bound)
    upper_bound = scale * (np.arange(n_timestamps_1) + horizontal_shift) \
        + vertical_shift
    upper_bound = np.round(upper_bound, 2)
    upper_bound = np.floor(upper_bound) + 1
    region = np.asarray([lower_bound, upper_bound]).astype('int64')
    region = _check_region(region, n_timestamps_1, n_timestamps_2)
    return region


def _dtw_sakoechiba(x=None, y=None, dist='square', window_size=0.1,
                    precomputed_cost=None, return_cost=False,
                    return_accumulated=False, return_path=False):
    """Dynamic Time Warping (DTW) distance with Sakoe-Chiba band constraint.

    The Sakoe-Chiba constraint region is a band centered around the main
    diagonal. Any cell whose distance to the diagonale is lower than a
    given value is valid for the path.

    Options
    -------
    window_size : float or int (default = 0.1)
        The window size above and below the diagonal.
        If float, `window_size` must be between 0 and
        1, and the actual window size will be computed as
        ``ceil(window_size * max((n_timestamps_1, n_timestamps_2) - 1))``.
        If int, `window_size` must be the largest temporal shift allowed.
        Each cell whose distance with the diagonal is lower than or equal to
        `window_size` becomes a valid cell for the path.

    References
    ----------
    .. [1] H. Sakoe and S. Chiba, "Dynamic programming algorithm optimization
           for spoken word recognition". IEEE Transactions on Acoustics,
           Speech, and Signal Processing, 26(1), 43-49 (1978).


    """
    x, y, precomputed_cost, n_timestamps_1, n_timestamps_2 = \
        _check_input_dtw(x, y, precomputed_cost, dist, method="sakoechiba")
    region = sakoe_chiba_band(n_timestamps_1, n_timestamps_2, window_size)
    cost_mat = _input_to_cost(x, y, dist, precomputed_cost, region=region)

    acc_cost_mat = accumulated_cost_matrix(cost_mat, region)
    dtw_dist = acc_cost_mat[-1, -1]
    if dist == 'square':
        dtw_dist = sqrt(dtw_dist)

    res = _return_results(dtw_dist, cost_mat, acc_cost_mat,
                          return_cost, return_accumulated, return_path)
    return res


def _get_itakura_slopes(n_timestamps_1, n_timestamps_2, max_slope):
    """Compute the slopes of the parallelogram bounds."""
    if not isinstance(n_timestamps_1, (int, np.integer)):
        raise TypeError("'n_timestamps_1' must be an integer.")
    else:
        if not n_timestamps_1 >= 2:
            raise ValueError("'n_timestamps_1' must be an integer greater than"
                             " or equal to 2.")

    if not isinstance(max_slope, (int, np.integer, float, np.floating)):
        raise TypeError("'max_slope' must be an integer or a float.")
    else:
        if not max_slope >= 1:
            raise ValueError("'max_slope' must be a number greater "
                             "than or equal to 1.")

    min_slope = 1 / max_slope
    scale_max = (n_timestamps_2 - 1) / (n_timestamps_1 - 2)
    max_slope *= scale_max
    max_slope = max(1., max_slope)

    scale_min = (n_timestamps_2 - 2) / (n_timestamps_1 - 1)

    min_slope *= scale_min
    min_slope = min(1., min_slope)
    return min_slope, max_slope


def itakura_parallelogram(n_timestamps_1, n_timestamps_2=None, max_slope=2.):
    """Compute the Itakura parallelogram.

    Parameters
    ----------
    n_timestamps_1 : int
        The size of the first time series.

    n_timestamps_2 : int (optional, default None)
        The size of the second time series. If None, set to `n_timestamps_1`.

    max_slope : float (default = 2.)
        Maximum slope for the parallelogram. Must be >= 1.

    Returns
    -------
    region : array, shape = (2, n_timestamps_1)
        Constraint region. The first row consists of the starting indices
        (included) and the second row consists of the ending indices (excluded)
        of the valid rows for each column.

    References
    ----------
    .. [1] F. Itakura, "Minimum prediction residual principle applied to
           speech recognition". IEEE Transactions on Acoustics,
           Speech, and Signal Processing, 23(1), 67–72 (1975).

    Examples
    --------
    >>> from pyts.metrics import itakura_parallelogram
    >>> print(itakura_parallelogram(5))
    [[0 1 1 2 4]
     [1 3 4 4 5]]

    """
    if n_timestamps_2 is None:
        n_timestamps_2 = n_timestamps_1
    min_slope_, max_slope_ = _get_itakura_slopes(
        n_timestamps_1, n_timestamps_2, max_slope)

    # Now we create the piecewise linear functions defining the parallelogram
    # lower_bound[0] = min_slope * x
    # lower_bound[1] = max_slope * (x - n_timestamps_1) + n_timestamps_2

    centered_scale = np.arange(n_timestamps_1) - n_timestamps_1 + 1
    lower_bound = np.empty((2, n_timestamps_1))
    lower_bound[0] = min_slope_ * np.arange(n_timestamps_1)
    lower_bound[1] = max_slope_ * centered_scale + n_timestamps_2 - 1

    # take the max of the lower linear funcs
    lower_bound = np.round(lower_bound, 2)
    lower_bound = np.ceil(np.max(lower_bound, axis=0))

    # upper_bound[0] = max_slope * x
    # upper_bound[1] = min_slope * (x - n_timestamps_1) + n_timestamps_2

    upper_bound = np.empty((2, n_timestamps_1))
    upper_bound[0] = max_slope_ * np.arange(n_timestamps_1) + 1
    upper_bound[1] = min_slope_ * centered_scale + n_timestamps_2

    # take the min of the upper linear funcs
    upper_bound = np.round(upper_bound, 2)
    upper_bound = np.floor(np.min(upper_bound, axis=0))

    # Little fix for max_slope = 1
    if max_slope == 1:
        if n_timestamps_2 > n_timestamps_1:
            upper_bound[:-1] = lower_bound[1:]
        else:
            upper_bound = lower_bound + 1

    region = np.asarray([lower_bound, upper_bound]).astype('int64')
    region = _check_region(region, n_timestamps_1, n_timestamps_2)
    return region


def _dtw_itakura(x=None, y=None, dist='square', max_slope=2.,
                 precomputed_cost=None, return_cost=False,
                 return_accumulated=False, return_path=False):
    """Dynamic Time Warping distance with Itakura parallelogram constraint.

    The Itakura constraint region is a parallelogram. Contrary to the
    Sakoe-Chiba band, whose width is constant, the Itakura parallelogram
    has a varying width, allowing for larger time shifts in the middle than
    at the first and last time points.

    Options
    -------
    max_slope : float (default = 2.)
        Maximum slope of the parallelogram.

    References
    ----------
    .. [1] F. Itakura, "Minimum prediction residual principle applied to
           speech recognition". IEEE Transactions on Acoustics,
           Speech, and Signal Processing, 23(1), 67–72 (1975).

    """
    x, y, precomputed_cost, n_timestamps_1, n_timestamps_2 = \
        _check_input_dtw(x, y, precomputed_cost, dist, method="itakura")
    region = itakura_parallelogram(n_timestamps_1, n_timestamps_2, max_slope)
    cost_mat = _input_to_cost(x, y, dist, precomputed_cost, region=region)
    acc_cost_mat = accumulated_cost_matrix(cost_mat, region)
    dtw_dist = acc_cost_mat[-1, -1]
    if dist == 'square':
        dtw_dist = sqrt(dtw_dist)

    res = _return_results(dtw_dist, cost_mat, acc_cost_mat,
                          return_cost, return_accumulated, return_path)
    return res


def _blurred_path_region(n_timestamps_1, n_timestamps_2, resolution_level,
                         n_timestamps_reduced_1, n_timestamps_reduced_2,
                         path, radius):
    path_length = path.shape[1]
    path_up = np.tile(path, radius)
    path_down = path_up.copy()
    path_left = path_up.copy()
    path_right = path_up.copy()

    for i in range(1, radius + 1):
        start = path_length * (i - 1)
        end = path_length * i
        path_up[0, start: end] += i
        path_down[0, start: end] -= i
        path_left[1, start: end] -= i
        path_right[1, start: end] += i

    path_radius = np.c_[path, path_up, path_down, path_left, path_right]
    path_radius[0] = np.clip(path_radius[0], 0, n_timestamps_reduced_1 - 1)
    path_radius[1] = np.clip(path_radius[1], 0, n_timestamps_reduced_2 - 1)

    region_reduced = np.empty((2, n_timestamps_reduced_1))
    for i in range(n_timestamps_reduced_1):
        arr = path_radius[1, path_radius[0] == i]
        min_, max_ = np.min(arr), np.max(arr)
        region_reduced[0, i] = min_ * resolution_level
        region_reduced[1, i] = (max_ + 1) * resolution_level

    region = np.repeat(region_reduced, resolution_level, axis=1)
    region = _check_region(region, n_timestamps_1, n_timestamps_2)

    return region.astype('int64')


def _multiscale_region(x, y, dist, resolution=2, radius=0):
    n_timestamps_1 = len(x)
    n_timestamps_2 = len(y)
    if not isinstance(resolution, (int, np.integer)):
        raise TypeError("'resolution' must be an integer.")
    if resolution < 1:
        raise ValueError("'resolution' must be a positive integer.")
    if not isinstance(radius, (int, np.integer)):
        raise TypeError("'radius' must be an integer.")
    if radius < 0:
        raise ValueError("'radius' must be a non-negative integer.")

    if resolution == 1:
        region = None
    else:
        remainder_1 = n_timestamps_1 % resolution
        remainder_2 = n_timestamps_2 % resolution

        if remainder_1 != 0:
            x_padded = np.append(x, np.repeat(x[-1], resolution - remainder_1))
            x_padded = x_padded.reshape(-1, resolution).mean(axis=1)
        else:
            x_padded = x.reshape(-1, resolution).mean(axis=1)
        if remainder_2 != 0:
            y_padded = np.append(y, np.repeat(y[-1], resolution - remainder_2))
            y_padded = y_padded.reshape(-1, resolution).mean(axis=1)
        else:
            y_padded = y.reshape(-1, resolution).mean(axis=1)
        cost_mat_res = cost_matrix(x_padded, y_padded, dist=dist, region=None)
        acc_cost_mat_res = accumulated_cost_matrix(cost_mat_res, region=None)
        path_res = _return_path(acc_cost_mat_res)
        region = _blurred_path_region(n_timestamps_1, n_timestamps_2,
                                      resolution, x_padded.size,
                                      y_padded.size, path_res,
                                      radius)
    return region


def _fast_region(x, y, dist, radius=0):
    if not isinstance(radius, (int, np.integer)):
        raise TypeError("'radius' must be an integer.")
    if radius < 0:
        raise ValueError("'radius' must be a non-negative integer.")
    n_timestamps_1, n_timestamps_2 = len(x), len(y)
    min_size = radius + 2
    region = None
    n_timestamps = min(n_timestamps_1, n_timestamps_2)
    if n_timestamps > min_size:
        n_recursions = ceil(log2(n_timestamps / min_size))
        for i in range(n_recursions):
            resolution = 2 ** (n_recursions - i)
            remainder_1 = n_timestamps_1 % resolution
            remainder_2 = n_timestamps_2 % resolution

            if remainder_1 != 0:
                x_padded = np.append(
                    x, np.repeat(x[-1], resolution - remainder_1)
                )
                x_padded = x_padded.reshape(-1, resolution).mean(axis=1)
            else:
                x_padded = x.reshape(-1, resolution).mean(axis=1)
            if remainder_2 != 0:
                y_padded = np.append(
                    y, np.repeat(y[-1], resolution - remainder_2)
                )
                y_padded = y_padded.reshape(-1, resolution).mean(axis=1)
            else:
                y_padded = y.reshape(-1, resolution).mean(axis=1)

            cost_mat_res = cost_matrix(x_padded, y_padded,
                                       dist=dist, region=region)
            acc_cost_mat_res = accumulated_cost_matrix(cost_mat_res,
                                                       region=region)
            path_res = _return_path(acc_cost_mat_res)
            n_timestamps_next_1 = ceil((2 * n_timestamps_1) / resolution)
            n_timestamps_next_2 = ceil((2 * n_timestamps_2) / resolution)

            region = _blurred_path_region(n_timestamps_next_1,
                                          n_timestamps_next_2,
                                          2, x_padded.size, y_padded.size,
                                          path_res, radius)

    return region


def _compute_region(n_timestamps_1, n_timestamps_2, method, dist,
                    x=None, y=None, **options):
    """Compute the region of feasible alignment paths."""

    if method == 'classic':
        region = None
    elif method == 'sakoechiba':
        region = sakoe_chiba_band(n_timestamps_1, n_timestamps_2, **options)
    elif method == 'itakura':
        region = itakura_parallelogram(n_timestamps_1, n_timestamps_2,
                                       **options)
    elif method == 'multiscale':
        region = _multiscale_region(x, y, dist, **options)
    elif method == 'fast':
        region = _fast_region(x, y, dist, **options)
    else:
        raise ValueError("'method' must be either 'classic', 'sakoechiba', "
                         "'itakura', 'multiscale' or 'fast'.")

    return region


def _dtw_multiscale(x, y, dist='square', resolution=2, radius=0,
                    return_cost=False, return_accumulated=False,
                    return_path=False):
    """Multiscale Dynamic Time Warping (DTW) distance.

    This version of DTW builds an adaptive constraint region. First both time
    series are downsampled, with the sizes of the original times series being
    divided by ``resolution``. Then the optimal path between the two
    downsampled time series is computed. Finally this optimal path is
    projected on the orginal scale and used as the constraint region.

    Options
    -------
    resolution : int (default = 2)
        The resolution level.

    radius : int (default = 0)
        The radius used to expand the constraint region. The optimal path
        computed at the resolution level is expanded with `radius` cells to the
        top, bottom, left and right of every cell belonging to the optimal
        path. It is computed at the resolution level.

    References
    ----------
    .. [1] M. Müller, H. Mattes and F. Kurth, "An efficient multiscale approach
           to audio synchronization". International Conference on Music
           Information Retrieval, 6(1), 192-197 (2006).

    """
    x, y, precomputed_cost, n_timestamps_1, n_timestamps_2 = \
        _check_input_dtw(x, y, precomputed_cost=None, dist=dist,
                         method="multiscale")

    region = _multiscale_region(x, y, dist, resolution=resolution,
                                radius=radius)
    cost_mat = cost_matrix(x, y, dist=dist, region=region)
    acc_cost_mat = accumulated_cost_matrix(cost_mat, region=region)
    dtw_dist = acc_cost_mat[-1, -1]
    if dist == 'square':
        dtw_dist = sqrt(dtw_dist)

    res = _return_results(dtw_dist, cost_mat, acc_cost_mat,
                          return_cost, return_accumulated, return_path)
    return res


def _dtw_fast(x, y, dist='square', radius=0, return_cost=False,
              return_accumulated=False, return_path=False):
    """Fast Dynamic Time Warping distance.

    This version of DTW builds an adaptive constraint region. The constraint
    region is created recursively by downsampling the time series, computing
    the optimal path and projecting it to a higher resolution. This process
    is repeated until the resolution matches the original resolution.

    Options
    -------
    radius : int (default = 0)
        The radius used to expand the constraint region. The optimal path
        computed at the resolution level is expanded with `radius` cells to the
        top, bottom, left and right of every cell belonging to the optimal
        path. It is computed at the resolution level.

    References
    ----------
    .. [1] S. Salvador ans P. Chan, "FastDTW: Toward Accurate Dynamic Time
           Warping in Linear Time and Space". KDD Workshop on Mining Temporal
           and Sequential Data, 70–80 (2004).

    """
    x, y, precomputed_cost, n_timestamps_1, n_timestamps_2 = \
        _check_input_dtw(x, y, precomputed_cost=None, dist=dist, method="fast")
    region = _fast_region(x, y, dist, radius=radius)
    cost_mat = cost_matrix(x, y, dist=dist, region=region)
    acc_cost_mat = accumulated_cost_matrix(cost_mat, region=region)
    dtw_dist = acc_cost_mat[-1, -1]
    if dist == 'square':
        dtw_dist = sqrt(dtw_dist)

    res = _return_results(dtw_dist, cost_mat, acc_cost_mat,
                          return_cost, return_accumulated, return_path)
    return res


def dtw(x=None, y=None, dist='square', method='classic', options=None,
        precomputed_cost=None, return_cost=False,
        return_accumulated=False, return_path=False):
    """Dynamic Time Warping (DTW) distance between two samples.

    Parameters
    ----------
    x : array-like, shape = (n_timestamps_1,)
        First array. Ignored if ``dist == 'precomputed'``.

    y : array-like, shape = (n_timestamps_2,)
        Second array. Ignored if ``dist == 'precomputed'``.

    dist : 'square', 'absolute', 'precomputed' or callable (default = 'square')
        Distance used. If 'square', the squared difference is used.
        If 'absolute', the absolute difference is used. If callable,
        it must be a function with a numba.njit() decorator that takes
        as input two numbers (two arguments) and returns a number.
        If 'precomputed', ``precomputed_cost`` must be the cost matrix and
        ``method`` must be 'classic', 'sakoechiba' or 'itakura'.

    method : str (default = 'classic')
        Method used. Should be one of

            - 'classic': :ref:`(see here) <metrics.dtw-classic>`
            - 'sakoechiba': :ref:`(see here) <metrics.dtw-sakoechiba>`
            - 'itakura': :ref:`(see here) <metrics.dtw-itakura>`
            - 'region': :ref:`(see here) <metrics.dtw-region>`
            - 'multiscale': :ref:`(see here) <metrics.dtw-multiscale>`
            - 'fast': :ref:`(see here) <metrics.dtw-fast>`

    options : None or dict (default = None)
        Dictionary of method options. Here is a quick summary of the available
        options for each method:

            - 'classic': None
            - 'sakoechiba': window_size (int or float)
            - 'itakura': max_slope (float)
            - 'region' : region (array-like)
            - 'multiscale': resolution (int) and radius (int)
            - 'fast': radius (int)

        For more information on these options, see :func:`show_options()`.

    precomputed_cost : array-like, shape = (n_timestamps_1, n_timestamps_2) \
            (default = None)
        Precomputed cost matrix between the time series.
        Ignored if ``dist != 'precomputed'``.

    return_cost : bool (default = False)
        If True, the cost matrix is returned.

    return_accumulated : bool (default = False)
        If True, the accumulated cost matrix is returned.

    return_path : bool (default = False)
        If True, the optimal path is returned.

    Returns
    -------
    dist : float
        The DTW distance between the two arrays.

    cost_mat : ndarray, shape = (n_timestamps_1, n_timestamps_2)
        Cost matrix. Only returned if ``return_cost=True``.

    acc_cost_mat : ndarray, shape = (n_timestamps_1, n_timestamps_2)
        Accumulated cost matrix. Only returned if ``return_accumulated=True``.

    path : ndarray, shape = (2, path_length)
        The optimal path along the cost matrix. The first row consists
        of the indices of the optimal path for x while the second row
        consists of the indices of the optimal path for y. Only returned
        if ``return_path=True``.

    Notes
    -----
    This section describes the available versions of Dynamic Time Warping (DTW)
    that can be selected by the `method` parameter.

    **Unconstrained path**

    Method :ref:`'classic'<metrics.dtw-classic>` computes the original DTW
    score between two time series with no constraint region: each cell is a
    valid cell to find the optimal path minimizing the total cost.

    **Global constraint region**

    Using no constraint region allows for very large distortion between both
    time series, which may be ill-suited for some cases. One possible solution
    to this issue is to use a global constraint region for the optimal path.
    A cell can be part of the optimal path if and only if this cell is in the
    constraint region. Another advantage of using a global constraint region is
    to decrease the computational complexity to find the optimal path.

    Method :ref:`'sakoechiba'<metrics.dtw-sakoechiba>` uses a band as the
    constraint region, allowing for a maximum fixed time shift at every time
    point.

    Method :ref:`'itakura'<metrics.dtw-itakura>` uses a parallelogram as
    the constraint region, allowing for time shifts of varying length: small
    time shifts at the beginning and at the end of the time series, larger
    time shifts in the middle.

    **Adaptative constraint region**

    One drawback of global constraint regions is that they do not depend on
    the time series and might be ill-suited for some time series. Adaptative
    constraint regions are regions that are specific to the two provided
    time series. However, their computational complexities are larger that
    the ones of global regions.

    Method :ref:`'multiscale'<metrics.dtw-multiscale>` downsamples both time
    series by a given factor, computes the optimal path with no constraint
    region for the downsampled time series, and projects the optimal path on
    the original scale to define the constraint region.

    Method :ref:`'fast'<metrics.dtw-fast>` can be seen as a recursive version
    of 'multiscale': the time series are downsampled so that the new time
    series are very small, the optimal path for these time series is computed
    and is projected at the scale that is two times larger. This process is
    repeated until the original scale of the time series is reached.

    References
    ----------
    .. [1] H. Sakoe and S. Chiba, "Dynamic programming algorithm optimization
           for spoken word recognition". IEEE Transactions on Acoustics,
           Speech, and Signal Processing, 26(1), 43-49 (1978).

    .. [2] F. Itakura, "Minimum prediction residual principle applied to
           speech recognition". IEEE Transactions on Acoustics,
           Speech, and Signal Processing, 23(1), 67–72 (1975).

    .. [3] M. Müller, H. Mattes and F. Kurth, "An efficient multiscale approach
           to audio synchronization". International Conference on Music
           Information Retrieval, 6(1), 192-197 (2006).

    .. [4] S. Salvador ans P. Chan, "FastDTW: Toward Accurate Dynamic Time
           Warping in Linear Time and Space". KDD Workshop on Mining Temporal
           and Sequential Data, 70–80 (2004).

    Examples
    --------
    >>> from pyts.metrics import dtw
    >>> x = [0, 1, 1]
    >>> y = [2, 0, 1]
    >>> dtw(x, y, method='sakoechiba', options={'window_size': 0.5})
    2.0

    """
    if options is None:
        options = dict()
    x, y, precomputed_cost, n_timestamps_1, n_timestamps_2 = \
        _check_input_dtw(x, y, precomputed_cost, dist, method)
    region = _compute_region(n_timestamps_1, n_timestamps_2, method, dist, x=x,
                             y=y, **options)
    cost_mat = _input_to_cost(x, y, dist, precomputed_cost, region)
    acc_cost_mat = accumulated_cost_matrix(cost_mat, region=region)
    dtw_dist = acc_cost_mat[-1, -1]
    if dist == 'square':
        dtw_dist = sqrt(dtw_dist)

    res = _return_results(dtw_dist, cost_mat, acc_cost_mat,
                          return_cost, return_accumulated, return_path)
    return res


def show_options(method=None, disp=True):
    """Show documentation for additional options of Dynamic Time Warping
    methods.

    These are method-specific options that can be supplied through the
    ``options`` dict.

    Parameters
    ----------
    method : None or str (default = None)
        If None, shows all methods. Otherwise, shows only the options for the
        specified method. If str, it must be either 'classic', 'sakoechiba',
        'itakura', 'region', 'multiscale' or 'fast'.

    disp : bool (default = True)
        Whether to print the result rather than returning it.

    Returns
    -------
    text : None or str
        Either None (if ``disp=True``) or the text string (if ``disp=False``).

    Notes
    -----
    The available methods are:

    - :ref:`'classic' <metrics.dtw-classic>`
    - :ref:`'sakoechiba' <metrics.dtw-sakoechiba>`
    - :ref:`'itakura' <metrics.dtw-itakura>`
    - :ref:`'region' <metrics.dtw-region>`
    - :ref:`'multiscale' <metrics.dtw-multiscale>`
    - :ref:`'fast' <metrics.dtw-fast>`

    """
    import textwrap

    def get_options(string):
        start = '    Options'
        end = '    References'
        str_lines = string.splitlines()
        start_idx = str_lines.index(start)
        end_idx = str_lines.index(end)
        doc = textwrap.dedent('\n'.join(str_lines[:start_idx])) + '\n'
        doc += textwrap.dedent('\n'.join(str_lines[start_idx:end_idx])) + '\n'
        return doc

    text = """\n"""
    if method is None:
        methods = ('classic', 'sakoechiba', 'itakura', 'region',
                   'multiscale', 'fast')
        for met in methods:
            text += show_options(met, disp=False)
    elif method == 'classic':
        text += "classic\n=======\n\n"
        text += get_options(_dtw_classic.__doc__)
    elif method == 'sakoechiba':
        text += "sakoechiba\n==========\n\n"
        text += get_options(_dtw_sakoechiba.__doc__)
    elif method == 'itakura':
        text += "itakura\n=======\n\n"
        text += get_options(_dtw_itakura.__doc__)
    elif method == 'region':
        text += "region\n=======\n\n"
        text += get_options(_dtw_itakura.__doc__)
    elif method == 'multiscale':
        text += "multiscale\n==========\n\n"
        text += get_options(_dtw_multiscale.__doc__)
    elif method == 'fast':
        text += "fast\n====\n\n"
        text += get_options(_dtw_fast.__doc__)
    else:
        raise ValueError("'method' must be either None, 'classic', "
                         "'sakoechiba', 'itakura', 'region', 'multiscale' "
                         "or 'fast'.")
    if disp:
        print(text)
        return
    else:
        return text



"""Code for Discrete Fourier Transform."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

import numpy as np
from sklearn.base import BaseEstimator
from sklearn.feature_selection import f_classif
from sklearn.utils.validation import check_array, check_is_fitted, check_X_y
from math import ceil
from warnings import warn
from ..base import UnivariateTransformerMixin
from ..preprocessing import StandardScaler


class DiscreteFourierTransform(BaseEstimator, UnivariateTransformerMixin):
    """Discrete Fourier Transform.

    Parameters
    ----------
    n_coefs : None, int or float (default = None)
        The number of Fourier coefficients to keep. If None, all the Fourier
        coeeficients are kept. If an integer, the ``n_coefs`` most significant
        Fourier coefficients are returned if ``anova=True``, otherwise the
        first ``n_coefs`` Fourier coefficients are returned. If a float, it
        represents a percentage of the size of each time series and must be
        between 0 and 1. The number of coefficients will be computed as
        ``ceil(n_coefs * (n_timestamps - 1))`` if ``drop_sum=True`` and
        ``ceil(n_coefs * n_timestamps)`` if ``drop_sum=False``.

    drop_sum : bool (default = False)
        If True, the first Fourier coefficient (i.e. the sum of the subseries)
        is dropped. Otherwise, it is kept.

    anova : bool (default = False)
        If True, the Fourier coefficient selection is done via a one-way
        ANOVA test. If False, the first Fourier coefficients are selected.

    norm_mean : bool (default = False)
        If True, center each time series before scaling.

    norm_std : bool (default = False)
        If True, scale each time series to unit variance.

    Attributes
    ----------
    support_ : array, shape = (n_coefs,)
        Indices of the kept Fourier coefficients.

    References
    ----------
    .. [1] P. Schäfer, and M. Högqvist, "SFA: A Symbolic Fourier Approximation
           and Index for Similarity Search in High Dimensional Datasets",
           International Conference on Extending Database Technology,
           15, 516-527 (2012).

    Examples
    --------
    >>> from pyts.approximation import DiscreteFourierTransform
    >>> from pyts.datasets import load_gunpoint
    >>> X, _, _, _ = load_gunpoint(return_X_y=True)
    >>> transformer = DiscreteFourierTransform(n_coefs=4)
    >>> X_new = transformer.fit_transform(X)
    >>> X_new.shape
    (50, 4)

    """

    def __init__(self, n_coefs=None, drop_sum=False, anova=False,
                 norm_mean=False, norm_std=False):
        self.n_coefs = n_coefs
        self.drop_sum = drop_sum
        self.anova = anova
        self.norm_mean = norm_mean
        self.norm_std = norm_std

    def fit(self, X, y=None):
        """Learn indices of the Fourier coefficients to keep.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_timestamps)
            Training vector.

        y : None or array-like, shape = (n_samples,) (default = None)
            Class labels for each data sample. Only used if ``anova=True``.

        Returns
        -------
        self : object

        """
        if self.anova:
            X, y = check_X_y(X, y, dtype='float64')
        else:
            X = check_array(X, dtype='float64')

        n_samples, n_timestamps = X.shape
        n_coefs = self._check_params(n_timestamps)
        if self.anova:
            ss = StandardScaler(self.norm_mean, self.norm_std)
            X = ss.fit_transform(X)
            X_fft = np.fft.rfft(X)
            X_fft = np.vstack([np.real(X_fft), np.imag(X_fft)])
            if n_timestamps % 2 == 0:
                X_fft = X_fft.reshape(n_samples, n_timestamps + 2, order='F')
                X_fft = np.c_[X_fft[:, 0], X_fft[:, 2:-1]]
            else:
                X_fft = X_fft.reshape(n_samples, n_timestamps + 1, order='F')
                X_fft = np.c_[X_fft[:, 0], X_fft[:, 2:]]
            if self.drop_sum:
                X_fft = X_fft[:, 1:]
            self.support_ = self._anova(X_fft, y, n_coefs, n_timestamps)
        else:
            self.support_ = np.arange(n_coefs)
        return self

    def transform(self, X):
        """Return the selected Fourier coefficients for each sample.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_timestamps)
            Input data.

        Returns
        -------
        X_new : array, shape (n_samples, n_coefs)
            The selected Fourier coefficients for each sample.

        """
        check_is_fitted(self, 'support_')
        X = check_array(X, dtype='float64')
        n_samples, n_timestamps = X.shape

        ss = StandardScaler(self.norm_mean, self.norm_std)
        X = ss.fit_transform(X)
        X_fft = np.fft.rfft(X)
        X_fft = np.vstack([np.real(X_fft), np.imag(X_fft)])
        if n_timestamps % 2 == 0:
            X_fft = X_fft.reshape(n_samples, n_timestamps + 2, order='F')
            X_fft = np.c_[X_fft[:, 0], X_fft[:, 2:-1]]
        else:
            X_fft = X_fft.reshape(n_samples, n_timestamps + 1, order='F')
            X_fft = np.c_[X_fft[:, 0], X_fft[:, 2:]]
        if self.drop_sum:
            X_fft = X_fft[:, 1:]
        return X_fft[:, self.support_]

    def fit_transform(self, X, y=None):
        """Learn and return the Fourier coeeficients to keep.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_timestamps)
            Training vector, where n_samples in the number of samples and
            n_features is the number of features.

        y : None or array-like, shape = (n_samples,) (default = None)
            Class labels for each data sample.

        Returns
        -------
        X_new : array, shape (n_samples, n_coefs)
            The selected Fourier coefficients for each sample.

        """
        if self.anova:
            X, y = check_X_y(X, y, dtype='float64')
        else:
            X = check_array(X, dtype='float64')
        n_samples, n_timestamps = X.shape
        n_coefs = self._check_params(n_timestamps)

        scaler = StandardScaler(self.norm_mean, self.norm_std)
        X = scaler.fit_transform(X)
        X_fft = np.fft.rfft(X)
        X_fft = np.vstack([np.real(X_fft), np.imag(X_fft)])
        if n_timestamps % 2 == 0:
            X_fft = X_fft.reshape(n_samples, n_timestamps + 2, order='F')
            X_fft = np.c_[X_fft[:, 0], X_fft[:, 2:-1]]
        else:
            X_fft = X_fft.reshape(n_samples, n_timestamps + 1, order='F')
            X_fft = np.c_[X_fft[:, 0], X_fft[:, 2:]]
        if self.drop_sum:
            X_fft = X_fft[:, 1:]
        if self.anova:
            self.support_ = self._anova(X_fft, y, n_coefs, n_timestamps)
        else:
            self.support_ = np.arange(n_coefs)
        return X_fft[:, self.support_]

    def _anova(self, X_fft, y, n_coefs, n_timestamps):
        if n_coefs < X_fft.shape[1]:
            non_constant = np.where(
                ~np.isclose(X_fft.var(axis=0), np.zeros_like(X_fft.shape[1]))
            )[0]
            if non_constant.size == 0:
                raise ValueError("All the Fourier coefficients are constant. "
                                 "Your input data is weirdly homogeneous.")
            elif non_constant.size < n_coefs:
                warn("The number of non constant Fourier coefficients ({0}) "
                     "is lower than the number of coefficients to keep ({1}). "
                     "The number of coefficients to keep is truncated to {2}"
                     ".".format(non_constant.size, n_coefs, non_constant.size))
                support = non_constant
            else:
                _, p = f_classif(X_fft[:, non_constant], y)
                support = non_constant[np.argsort(p)[:n_coefs]]
        else:
            support = np.arange(n_coefs)
        return support

    def _check_params(self, n_timestamps):
        if not ((isinstance(self.n_coefs,
                            (int, np.integer, float, np.floating)))
                or (self.n_coefs is None)):
            raise TypeError("'n_coefs' must be None, an integer or a float.")
        if isinstance(self.n_coefs, (int, np.integer)):
            if self.drop_sum and not (1 <= self.n_coefs <= n_timestamps - 1):
                raise ValueError(
                    "If 'n_coefs' is an integer, it must be greater than or "
                    "equal to 1 and lower than or equal to (n_timestamps - 1) "
                    "if 'drop_sum=True'."
                )
            if not self.drop_sum and not (1 <= self.n_coefs <= n_timestamps):
                raise ValueError(
                    "If 'n_coefs' is an integer, it must be greater than or "
                    "equal to 1 and lower than or equal to n_timestamps "
                    "if 'drop_sum=False'."
                )
            n_coefs = self.n_coefs
        elif isinstance(self.n_coefs, (float, np.floating)):
            if not 0 < self.n_coefs <= 1:
                raise ValueError(
                    "If 'n_coefs' is a float, it must be greater "
                    "than 0 and lower than or equal to 1."
                )
            if self.drop_sum:
                n_coefs = ceil(self.n_coefs * (n_timestamps - 1))
            else:
                n_coefs = ceil(self.n_coefs * n_timestamps)
        else:
            n_coefs = (n_timestamps - 1) if self.drop_sum else n_timestamps
        return n_coefs



"""Testing for Lower Bounds of Dynamic Time Warping."""

import numpy as np
import pytest
import re
from math import sqrt

from pyts.metrics.lower_bounds import (
    _lower_bound_yi_x_y, _lower_bound_yi_X_Y, _warping_envelope, _clip
)
from pyts.metrics import (lower_bound_improved, lower_bound_keogh,
                          lower_bound_kim, lower_bound_yi)
from pyts.metrics import dtw, sakoe_chiba_band
from pyts.metrics.dtw import _dtw_sakoechiba
from sklearn.metrics import pairwise_distances


@pytest.mark.parametrize(
    'X, Y, err_msg',
    [([[1, 1]], [[1]], "Found input variables with inconsistent numbers of "
                       "timestamps: [2, 1]"),
     ([[1]], [[1, 2]], "Found input variables with inconsistent numbers of "
                       "timestamps: [1, 2]"),
     ([[3, 1, 1]], [[1]], "Found input variables with inconsistent numbers of "
                          "timestamps: [3, 1]")]
)
def test_check_consistent_lengths(X, Y, err_msg):
    """Test 'lower_bound_yi' parameter validation."""
    with pytest.raises(ValueError, match=re.escape(err_msg)):
        lower_bound_yi(X, Y)


@pytest.mark.parametrize(
    'x, y, float_desired',
    [([1, 5, 3, 2, 8], [3, 2, 5, 4, 5], sqrt(10)),
     ([4, 5, 3, 6, 8], [1, 2, 0, 1, 2], sqrt(66)),
     ([4, 5, 3, 6, 8], [1, 5, 3, 2, 1], sqrt(19)),
     ([3, 2, 5, 4, 5], [1, 5, 3, 2, 8], sqrt(10)),
     ([1, 2, 0, 1, 2], [4, 5, 3, 6, 8], sqrt(66)),
     ([1, 5, 3, 2, 1], [4, 5, 3, 6, 8], sqrt(19))]
)
def test_lower_bound_yi_x_y(x, y, float_desired):
    """Test '_lower_bound_yi_x_y' function."""
    x, y = np.asarray(x), np.asarray(y)
    float_actual = _lower_bound_yi_x_y(x, min(x), max(x), y, min(y), max(y))
    np.testing.assert_allclose(float_actual, float_desired, atol=1e-5, rtol=0)


@pytest.mark.parametrize(
    'X, Y',
    [([[3, 5, 1, 4, 6], [3, 8, 2, 4, 2]],
      [[4, 5, 9, 2, 3], [4, 3, 5, 2, 3], [5, 9, 3, 3, 4]])]
)
def test_lower_bound_yi_X_Y(X, Y):
    """Test '_lower_bound_yi_X_Y' function."""
    X, Y = np.asarray(X), np.asarray(Y)
    arr_actual = _lower_bound_yi_X_Y(X, X.min(axis=1), X.max(axis=1),
                                     Y, Y.min(axis=1), Y.max(axis=1))
    arr_desired = np.empty((2, 3))
    for i in range(2):
        for j in range(3):
            arr_desired[i, j] = _lower_bound_yi_x_y(
                X[i], X[i].min(), X[i].max(),
                Y[j], Y[j].min(), Y[j].max()
            )
    np.testing.assert_allclose(arr_actual, arr_desired, atol=1e-5, rtol=0)


@pytest.mark.parametrize(
    'X_train, X_test, arr_desired',
    [([[5, 4, 3, 2, 1], [1, 8, 4, 3, 2], [6, 3, 5, 4, 7]],
      [[2, 1, 8, 4, 5]],
      np.sqrt([[9, 0, 6]]))]
)
def test_actual_results_lower_bound_yi(X_train, X_test, arr_desired):
    """Test that the actual results are the expected ones."""
    arr_actual = lower_bound_yi(X_train, X_test)
    np.testing.assert_allclose(arr_actual, arr_desired, atol=1e-5, rtol=0)


@pytest.mark.parametrize(
    'X_train, X_test, arr_desired',
    [([[2, 1, 8, 4, 5], [1, 2, 3, 4, 5]],
      [[5, 4, 3, 2, 1], [1, 8, 4, 3, 2], [6, 3, 5, 4, 7]],
      [[4, 4], [3, 3], [4, 5]])]
)
def test_actual_results_lower_bound_kim(X_train, X_test, arr_desired):
    """Test that the actual results are the expected ones."""
    arr_actual = lower_bound_kim(X_train, X_test)
    np.testing.assert_array_equal(arr_actual, arr_desired)


@pytest.mark.parametrize(
    'X, region, err_msg',
    [([0, 1], [[0, 1], [0, 1]], "X must be a two- or three-dimensional."),
     ([[[[0, 1, 2]]]], [[0, 1], [0, 1]],
      "X must be a two- or three-dimensional.")]
)
def test_parameter_check_warping_envelope(X, region, err_msg):
    """Test '_warping_envelope' parameter validation.."""
    with pytest.raises(ValueError, match=err_msg):
        _warping_envelope(X, region)


@pytest.mark.parametrize(
    'X, region, lower_desired, upper_desired',
    [([[0, 1, 2, 3], [1, 5, 3, -1]], [[0, 0, 1, 2], [2, 3, 4, 4]],
      [[0, 0, 1, 2], [1, 1, -1, -1]], [[1, 2, 3, 3], [5, 5, 5, 3]]),
     ([[[0, 1, 2, 3], [1, 5, 3, -1]]], [[0, 0, 1, 2], [2, 3, 4, 4]],
      [[[0, 0, 1, 2], [1, 1, -1, -1]]], [[[1, 2, 3, 3], [5, 5, 5, 3]]])]
)
def test_actual_results_warping_envelope(X, region,
                                         lower_desired, upper_desired):
    """Test that the actual results are the expected ones."""
    lower_actual, upper_actual = _warping_envelope(X, region)
    np.testing.assert_array_equal(lower_actual, lower_desired)
    np.testing.assert_array_equal(upper_actual, upper_desired)


@pytest.mark.parametrize(
    'X, lower, upper, err_msg',
    [([[0], [1]], [6], [[1], [1]],
      "'lower' must be two- or three-dimensional."),
     ([[0], [1]], [[[[6]]]], [[1], [1]],
      "'lower' must be two- or three-dimensional."),
     ([[0], [1]], [[[6]]], [[1], [1]],
      "'lower' and 'upper' must have the same shape ((1, 1, 1) != (2, 1))")]
)
def test_parameter_check_clip(X, lower, upper, err_msg):
    """Test '_clip' parameter validation.."""
    with pytest.raises(ValueError, match=re.escape(err_msg)):
        _clip(X, lower, upper)


@pytest.mark.parametrize(
    'X, lower, upper, arr_desired',
    [([[0, 1, 2, 3], [1, 5, 3, -1]],
      [[0, 3, 3, 3], [-1, 2, 4, 6]],
      [[1, 5, 4, 6], [1, 3, 6, 8]],
      [[[0, 3, 3, 3], [0, 2, 4, 6]], [[1, 5, 3, 3], [1, 3, 4, 6]]]),

     ([[0, 1, 2, 3], [1, 5, 3, -1]],
      [[[0, 3, 3, 3], [-1, 2, 4, 6]]],
      [[[1, 5, 4, 6], [1, 3, 6, 8]]],
      [[[0, 3, 3, 3]], [[1, 3, 4, 6]]])]
)
def test_actual_results_clip(X, lower, upper, arr_desired):
    """Test that the actual results are the expected ones."""
    arr_actual = _clip(X, lower, upper)
    np.testing.assert_array_equal(arr_actual, arr_desired)


def test_actual_results_lower_bound_keogh():
    """Test that the actual results are the expected ones."""
    # Toy dataset
    X_train = np.asarray([[0, 1, 2, 3],
                          [1, 2, 3, 4]])
    X_test = np.asarray([[0, 2.5, 3.5, 6]])

    # Region = Sakoe-Chiba band (w=0)
    region = [[0, 1, 2, 3],
              [1, 2, 3, 4]]
    arr_actual = lower_bound_keogh(X_train, X_test, region)
    arr_desired = np.sqrt(np.sum(
        (X_test[:, None, :] - X_train[None, :, :]) ** 2, axis=-1
    ))
    np.testing.assert_allclose(arr_actual, arr_desired, atol=1e-5, rtol=0)

    # Region = Sakoe-Chiba band (w=1)
    region_window = [[0, 0, 1, 2],
                     [2, 3, 4, 4]]
    arr_actual_window = lower_bound_keogh(X_train, X_test, region_window)
    # lower = [[0, 0, 1, 2], [1, 1, 2, 3]]
    # upper = [[1, 2, 3, 3], [2, 3, 4, 4]]
    # X_proj = [[0, 2, 3, 3], [1, 2.5, 3.5, 4]]
    # LB_Keogh = [[sqrt(0.25 + 0.25 + 9), sqrt(1 + 4)]]
    arr_desired_window = np.sqrt([[9.5, 5]])
    np.testing.assert_allclose(arr_actual_window, arr_desired_window,
                               atol=1e-5, rtol=0)


def test_actual_results_lower_bound_improved():
    """Test that the actual results are the expected ones."""
    # Toy dataset
    X_train = np.asarray([[0, 1, 2, 3],
                          [1, 2, 3, 4]])
    X_test = np.asarray([[0, 2.5, 3.5, 3.3]])

    # Region = Sakoe-Chiba band (w=0)
    region = [[0, 1, 2, 3],
              [1, 2, 3, 4]]
    arr_actual = lower_bound_improved(X_train, X_test, region)
    arr_desired = np.sqrt(np.sum(
        (X_test[:, None, :] - X_train[None, :, :]) ** 2, axis=-1
    ))
    np.testing.assert_allclose(arr_actual, arr_desired, atol=1e-5, rtol=0)

    # Region = Sakoe-Chiba band (w=1)
    region_window = [[0, 0, 1, 2],
                     [2, 3, 4, 4]]
    arr_actual_window = lower_bound_improved(X_train, X_test, region_window)
    # lower_train = [[0, 0, 1, 2], [1, 1, 2, 3]
    # upper_train = [[1, 2, 3, 3], [2, 3, 4, 4]]
    # X_test_proj = [[0, 2, 3, 3], [1, 2.5, 3.5, 3.3]
    # LB_Keogh^2 = [[0.25 + 0.25 + 0.09, 1]] = [[0.59, 1]]
    # lower_test = [[0, 0, 2, 3], [1, 1, 2.5, 3.3]]
    # upper_test = [[2, 3, 3, 3], [2.5, 3.5, 3.5, 3.5]]
    # X_train_proj = [[0, 1, 2, 3], [1, 2, 3, 3.5]]
    # LB_Improved^2 = [[0, 0.25]]
    arr_desired_window = np.sqrt([[0.59 + 0, 1 + 0.25]])
    np.testing.assert_allclose(arr_actual_window, arr_desired_window,
                               atol=1e-5, rtol=0)


def test_lower_bounds_inequalities():
    """Test that the expected inequalities are verified."""
    # Toy dataset
    rng = np.random.RandomState(42)
    n_samples_train, n_samples_test, n_timestamps = 20, 30, 60
    window_size = 0.1
    X_train = rng.randn(n_samples_train, n_timestamps)
    X_test = rng.randn(n_samples_test, n_timestamps)

    # DTW
    X_dtw = pairwise_distances(X_test, X_train, dtw)
    region = sakoe_chiba_band(n_timestamps, window_size=window_size)
    X_dtw_window = pairwise_distances(X_test, X_train, _dtw_sakoechiba,
                                      window_size=window_size)

    # Lower bounds
    lb_yi = lower_bound_yi(X_train, X_test)
    lb_kim = lower_bound_kim(X_train, X_test)
    lb_keogh = lower_bound_keogh(X_train, X_test, region)
    lb_improved = lower_bound_improved(X_train, X_test, region)

    # Sanity check
    EPS = 1e-8
    np.testing.assert_array_less(lb_yi, X_dtw + EPS)
    np.testing.assert_array_less(lb_kim, X_dtw + EPS)
    np.testing.assert_array_less(lb_keogh, X_dtw_window + EPS)
    np.testing.assert_array_less(lb_improved, X_dtw_window + EPS)
    np.testing.assert_array_less(lb_keogh, lb_improved + EPS)



"""Testing for Recurrence Plot."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

import numpy as np
import pytest
import re
from pyts.image.recurrence import _trajectories
from pyts.image import RecurrencePlot


X = np.asarray([[0, 1, 1, 0, 2, 3],
                [1, 0, 3, 2, 1, 0]])


@pytest.mark.parametrize(
    'params, arr_desired',
    [({'dimension': 3, 'time_delay': 1},
      np.asarray([[[0, 1, 1], [1, 1, 0], [1, 0, 2], [0, 2, 3]],
                  [[1, 0, 3], [0, 3, 2], [3, 2, 1], [2, 1, 0]]])),

     ({'dimension': 2, 'time_delay': 2},
      np.asarray([[[0, 1], [1, 0], [1, 2], [0, 3]],
                  [[1, 3], [0, 2], [3, 1], [2, 0]]])),

     ({'dimension': 3, 'time_delay': 2},
      np.asarray([[[0, 1, 2], [1, 0, 3]], [[1, 3, 1], [0, 2, 0]]])),

     ({'dimension': 1, 'time_delay': 1},
      np.asarray([[[0], [1], [1], [0], [2], [3]],
                  [[1], [0], [3], [2], [1], [0]]])),

     ({'dimension': 1, 'time_delay': 5},
      np.asarray([[[0], [1], [1], [0], [2], [3]],
                  [[1], [0], [3], [2], [1], [0]]]))]
)
def test_actual_results_trajectories(params, arr_desired):
    """Test that the actual results are the expected ones."""
    arr_actual = _trajectories(X, **params)
    np.testing.assert_allclose(arr_actual, arr_desired, atol=1e-5, rtol=0.)


@pytest.mark.parametrize(
    'params, error, err_msg',
    [({'dimension': '3'}, TypeError,
      "'dimension' must be an integer or a float."),

     ({'time_delay': '3'}, TypeError,
      "'time_delay' must be an integer or a float."),

     ({'threshold': '3'}, TypeError,
      "'threshold' must be either None, 'point', 'distance', a float or an "
      "integer."),

     ({'percentage': '3'}, TypeError,
      "'percentage' must be a float or an integer."),

     ({'dimension': 10}, ValueError,
      "If 'dimension' is an integer, it must be greater than or equal to 1 "
      "and lower than or equal to n_timestamps (got 10)."),

     ({'dimension': 2.}, ValueError,
      "If 'dimension' is a float, it must be greater than 0 and lower than or "
      "equal to 1 (got {0}).".format(2.)),

     ({'time_delay': 12}, ValueError,
      "If 'time_delay' is an integer, it must be greater than or equal to 1 "
      "and lower than or equal to n_timestamps (got 12)."),

     ({'time_delay': 2.}, ValueError,
      "If 'time_delay' is a float, it must be greater than 0 and lower than "
      "or equal to 1 (got {0}).".format(2.)),

     ({'time_delay': 3, 'dimension': 3}, ValueError,
      "The number of trajectories must be positive. Consider trying with "
      "smaller values for 'dimension' and 'time_delay'."),

     ({'threshold': -1}, ValueError,
      "If 'threshold' is a float or an integer, it must be greater than or "
      "equal to 0."),

     ({'percentage': 200, 'threshold': 'point'}, ValueError,
      "'percentage' must be between 0 and 100.")]
)
def test_parameter_check(params, error, err_msg):
    """Test parameter validation."""
    recurrence = RecurrencePlot(**params)
    with pytest.raises(error, match=re.escape(err_msg)):
        recurrence.transform(X)


@pytest.mark.parametrize(
    'params, arr_desired',
    [({'dimension': 3, 'time_delay': 1},
     [[[0, 2, 3, 5], [2, 0, 5, 11], [3, 5, 0, 6], [5, 11, 6, 0]],
      [[0, 11, 12, 11], [11, 0, 11, 12], [12, 11, 0, 3], [11, 12, 3, 0]]]),

     ({'dimension': 2, 'time_delay': 2},
      [[[0, 2, 2, 4], [2, 0, 4, 10], [2, 4, 0, 2], [4, 10, 2, 0]],
       [[0, 2, 8, 10], [2, 0, 10, 8], [8, 10, 0, 2], [10, 8, 2, 0]]]),

     ({'dimension': 3, 'time_delay': 2},
      [[[0, 3], [3, 0]], [[0, 3], [3, 0]]]),

     ({'dimension': 0.5, 'time_delay': 0.3},
      [[[0, 3], [3, 0]], [[0, 3], [3, 0]]]),

     ({'dimension': 0.5, 'time_delay': 0.3,
       'threshold': 'point', 'percentage': 50},
      [[[1, 0], [0, 1]], [[1, 0], [0, 1]]]),

     ({'dimension': 0.5, 'time_delay': 0.3,
       'threshold': 'distance', 'percentage': 50},
      [[[1, 0], [0, 1]], [[1, 0], [0, 1]]]),

     ({'dimension': 1, 'time_delay': 1},
      [[[0, 1, 1, 0, 4, 9], [1, 0, 0, 1, 1, 4], [1, 0, 0, 1, 1, 4],
        [0, 1, 1, 0, 4, 9], [4, 1, 1, 4, 0, 1], [9, 4, 4, 9, 1, 0]],
       [[0, 1, 4, 1, 0, 1], [1, 0, 9, 4, 1, 0], [4, 9, 0, 1, 4, 9],
        [1, 4, 1, 0, 1, 4], [0, 1, 4, 1, 0, 1], [1, 0, 9, 4, 1, 0]]]),

     ({'dimension': 1, 'time_delay': 5},
      [[[0, 1, 1, 0, 4, 9], [1, 0, 0, 1, 1, 4], [1, 0, 0, 1, 1, 4],
        [0, 1, 1, 0, 4, 9], [4, 1, 1, 4, 0, 1], [9, 4, 4, 9, 1, 0]],
       [[0, 1, 4, 1, 0, 1], [1, 0, 9, 4, 1, 0], [4, 9, 0, 1, 4, 9],
        [1, 4, 1, 0, 1, 4], [0, 1, 4, 1, 0, 1], [1, 0, 9, 4, 1, 0]]])]
)
def test_actual_results_recurrence_plot(params, arr_desired):
    """Test that the actual results are the expected ones."""
    recurrence = RecurrencePlot(**params)
    arr_actual = recurrence.fit_transform(X) ** 2
    np.testing.assert_allclose(arr_actual, arr_desired, atol=1e-5, rtol=0.)


def test_flatten():
    """Test the 'flatten' parameter."""
    arr_false = RecurrencePlot().transform(X).reshape(2, -1)
    arr_true = RecurrencePlot(flatten=True).transform(X)
    np.testing.assert_allclose(arr_false, arr_true, atol=1e-5, rtol=0.)



"""Testing for Word ExtrAction for time SEries cLassification."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

import numpy as np
import pytest
import re
from scipy.sparse import csr_matrix, hstack
from sklearn.feature_extraction.text import CountVectorizer
from sklearn.feature_selection import chi2
from pyts.approximation import SymbolicFourierApproximation
from pyts.transformation import WEASEL


n_samples, n_timestamps, n_classes = 8, 200, 2
rng = np.random.RandomState(42)
X = rng.randn(n_samples, n_timestamps)
y = rng.randint(n_classes, size=n_samples)


@pytest.mark.parametrize(
    'params, error, err_msg',
    [({'word_size': "3"}, TypeError, "'word_size' must be an integer."),

     ({'window_sizes': {}}, TypeError, "'window_sizes' must be array-like."),

     ({'window_steps': "3"}, TypeError,
      "'window_steps' must be None or array-like."),

     ({'chi2_threshold': "3"}, TypeError,
      "'chi2_threshold' must be a float or an integer."),

     ({'word_size': 0}, ValueError, "'word_size' must be a positive integer."),

     ({'window_sizes': np.ones((2, 4))}, ValueError,
      "'window_sizes' must be one-dimensional."),

     ({'window_sizes': ['a', 'b', 'c']}, ValueError,
      "The elements of 'window_sizes' must be integers or floats."),

     ({'window_sizes': [0.5, 2.]}, ValueError,
      "If the elements of 'window_sizes' are floats, they all must be greater "
      "than 0 and lower than or equal to 1."),

     ({'window_sizes': [300]}, ValueError,
      "All the elements in 'window_sizes' must be lower than or equal to "
      "n_timestamps."),

     ({'word_size': 4, 'window_sizes': [4, 6], 'drop_sum': True}, ValueError,
      "If 'drop_sum=True', 'word_size' must be lower than the minimum value "
      "in 'window_sizes'."),

     ({'word_size': 5, 'window_sizes': [4, 6], 'drop_sum': False}, ValueError,
      "If 'drop_sum=False', 'word_size' must be lower than or equal to the "
      "minimum value in 'window_sizes'."),

     ({'window_steps': np.ones((2, 4))}, ValueError,
      "'window_steps' must be one-dimensional."),

     ({'window_sizes': [8, 10], 'window_steps': [1, 2, 3]}, ValueError,
      "If 'window_steps' is not None, it must have the same size as "
      "'window_sizes'."),

     ({'window_sizes': [8, 10], 'window_steps': ['a', 'b']}, ValueError,
      "If 'window_steps' is not None, the elements of 'window_steps' must be "
      "integers or floats."),

     ({'window_sizes': [8, 10], 'window_steps': [0.5, 2.]}, ValueError,
      "If the elements of 'window_steps' are floats, they all must be greater "
      "than 0 and lower than or equal to 1."),

     ({'window_sizes': [8], 'window_steps': [0]}, ValueError,
      "All the elements in 'window_steps' must be greater than or equal to 1 "
      "and lower than or equal to n_timestamps."),

     ({'window_sizes': [8], 'window_steps': [250]}, ValueError,
      "All the elements in 'window_steps' must be greater than or equal to 1 "
      "and lower than or equal to n_timestamps."),

     ({'chi2_threshold': -1}, ValueError,
      "'chi2_threshold' must be positive.")]
)
def test_parameter_check(params, error, err_msg):
    """Test parameter validation."""
    weasel = WEASEL(**params)
    with pytest.raises(error, match=re.escape(err_msg)):
        weasel.fit(X, y)


@pytest.mark.parametrize(
    'sparse, instance', [(True, csr_matrix), (False, np.ndarray)])
def test_sparse_dense(sparse, instance):
    """Test that the expected type is returned."""
    weasel = WEASEL(strategy='quantile', sparse=sparse)
    assert isinstance(weasel.fit(X, y).transform(X), instance)
    assert isinstance(weasel.fit_transform(X, y), instance)


@pytest.mark.parametrize(
    'params',
    [{'window_sizes': [0.1, 0.2, 0.3], 'window_steps': [0.1, 0.2, 0.3]},
     {'window_sizes': [0.1], 'window_steps': [0.2]}]
)
def test_fit_transform(params):
    """Test that 'fit_transform' and 'fit' then 'transform' yield same res."""
    weasel_1 = WEASEL(**params, sparse=False)
    weasel_2 = WEASEL(**params, sparse=False)

    X_fit_transform = weasel_1.fit_transform(X, y)
    X_fit_then_transform = weasel_2.fit(X, y).transform(X)

    # Test that the transformation are identical
    np.testing.assert_allclose(X_fit_transform, X_fit_then_transform,
                               atol=1e-5, rtol=0.)

    # Test that the vocabulary_ attributes are identical
    assert weasel_1.vocabulary_ == weasel_2.vocabulary_


def test_accurate_results():
    """Test that the actual results are the expected ones."""
    X_features = csr_matrix((n_samples, 0), dtype=np.int64)
    vocabulary_ = {}

    weasel = WEASEL(
        word_size=4, n_bins=3, window_sizes=[5, 10],
        window_steps=None, anova=True, drop_sum=True, norm_mean=True,
        norm_std=True, strategy='entropy', chi2_threshold=2, alphabet=None
    )

    for window_size, n_windows in zip([5, 10], [40, 20]):
        X_windowed = X.reshape(n_samples, n_windows, window_size)
        X_windowed = X_windowed.reshape(n_samples * n_windows, window_size)

        sfa = SymbolicFourierApproximation(
            n_coefs=4, drop_sum=True, anova=True, norm_mean=True,
            norm_std=True, n_bins=3, strategy='entropy', alphabet=None
        )
        y_repeated = np.repeat(y, n_windows)
        X_sfa = sfa.fit_transform(X_windowed, y_repeated)
        X_word = np.asarray([''.join(X_sfa[i])
                             for i in range((n_samples * n_windows))])
        X_word = X_word.reshape(n_samples, n_windows)
        X_bow = np.asarray([' '.join(X_word[i]) for i in range(n_samples)])

        vectorizer = CountVectorizer(ngram_range=(1, 2))
        X_counts = vectorizer.fit_transform(X_bow)
        chi2_statistics, _ = chi2(X_counts, y)
        relevant_features = np.where(
            chi2_statistics > 2)[0]
        X_features = hstack([X_features, X_counts[:, relevant_features]])

        old_length_vocab = len(vocabulary_)
        vocabulary = {value: key
                      for (key, value) in vectorizer.vocabulary_.items()}
        for i, idx in enumerate(relevant_features):
            vocabulary_[i + old_length_vocab] = \
                str(window_size) + " " + vocabulary[idx]

    arr_desired = X_features.toarray()

    # Accuracte results for fit followed by transform
    arr_actual_1 = weasel.fit_transform(X, y).toarray()
    np.testing.assert_allclose(arr_actual_1, arr_desired, atol=1e-5, rtol=0)
    assert weasel.vocabulary_ == vocabulary_

    # Accuracte results for fit_transform
    arr_actual_2 = weasel.fit(X, y).transform(X).toarray()
    np.testing.assert_allclose(arr_actual_2, arr_desired, atol=1e-5, rtol=0)
    assert weasel.vocabulary_ == vocabulary_



"""Testing for SAX-VSM."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

import numpy as np
from math import log
from sklearn.metrics.pairwise import cosine_similarity
from pyts.classification import SAXVSM


X = [[0, 0, 0, 1, 0, 0, 1, 1, 1],
     [0, 1, 1, 1, 0, 0, 1, 1, 1],
     [0, 0, 0, 1, 0, 0, 0, 1, 0]]
y = [0, 0, 1]


def test_actual_results_strategy_uniform():
    """Test that the actual results are the expected ones."""
    # Data
    X = [[0, 0, 0, 1, 0, 0, 1, 1, 1],
         [0, 1, 1, 1, 0, 0, 1, 1, 1],
         [0, 0, 0, 1, 0, 0, 0, 1, 0]]
    y = [0, 0, 1]

    clf = SAXVSM(window_size=4, word_size=4, n_bins=2, strategy='uniform',
                 numerosity_reduction=False, sublinear_tf=False)
    decision_function_actual = clf.fit(X, y).decision_function(X)

    # X_bow = ["aaab aaba abaa baab aabb abbb",
    #          "abbb bbba bbaa baab aabb abbb",
    #          "aaab aaba abaa baaa aaab aaba"]

    assert clf.vocabulary_ == {0: 'aaab', 1: 'aaba', 2: 'aabb', 3: 'abaa',
                               4: 'abbb', 5: 'baaa', 6: 'baab', 7: 'bbaa',
                               8: 'bbba'}

    freq = np.asarray([[1, 1, 1, 1, 1, 0, 1, 0, 0],
                       [0, 0, 1, 0, 2, 0, 1, 1, 1],
                       [2, 2, 0, 1, 0, 1, 0, 0, 0]])
    tf = np.asarray([[1, 1, 2, 1, 3, 0, 2, 1, 1],
                     [2, 2, 0, 1, 0, 1, 0, 0, 0]])
    idf = np.asarray([1, 1, log(2) + 1, 1, log(2) + 1, log(2) + 1, log(2) + 1,
                      log(2) + 1, log(2) + 1])
    decision_function_desired = cosine_similarity(freq, tf * idf[None, :])
    np.testing.assert_allclose(decision_function_actual,
                               decision_function_desired, atol=1e-5, rtol=0.)

    pred_actual = clf.predict(X)
    pred_desired = cosine_similarity(freq, tf * idf[None, :]).argmax(axis=1)
    np.testing.assert_array_equal(pred_actual, pred_desired)


def test_actual_results_strategy_quantile():
    """Test that the actual results are the expected ones."""
    # Data
    X = [[0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9],
         [0.0, 0.3, 0.2, 0.4, 0.1, 0.5, 0.6, 0.7, 0.8, 0.9],
         [0.0, 0.9, 0.1, 0.8, 0.2, 0.7, 0.3, 0.6, 0.4, 0.5]]
    y = [0, 0, 1]

    clf = SAXVSM(window_size=4, word_size=4, n_bins=2, strategy='quantile',
                 numerosity_reduction=False, sublinear_tf=False)
    decision_function_actual = clf.fit(X, y).decision_function(X)

    # X_bow = ["aabb aabb aabb aabb aabb aabb aabb",
    #          "abab baba abab aabb aabb aabb aabb",
    #          "abab baba abab baba abab baba abab"]

    assert clf.vocabulary_ == {0: 'aabb', 1: 'abab', 2: 'baba'}

    freq = np.asarray([[7, 0, 0],
                       [4, 2, 1],
                       [0, 4, 3]])
    tf = np.asarray([[11, 2, 1],
                     [0, 4, 3]])
    idf = np.asarray([log(2) + 1, 1, 1])
    decision_function_desired = cosine_similarity(freq, tf * idf[None, :])
    np.testing.assert_allclose(decision_function_actual,
                               decision_function_desired, atol=1e-5, rtol=0.)

    pred_actual = clf.fit(X, y).predict(X)
    pred_desired = cosine_similarity(freq, tf * idf[None, :]).argmax(axis=1)
    np.testing.assert_array_equal(pred_actual, pred_desired)

    # Without IDF weighting
    clf = SAXVSM(window_size=4, word_size=4, n_bins=2, strategy='quantile',
                 numerosity_reduction=False, sublinear_tf=False, use_idf=False)
    decision_function_actual = clf.fit(X, y).decision_function(X)

    decision_function_desired = cosine_similarity(freq, tf)
    np.testing.assert_allclose(decision_function_actual,
                               decision_function_desired, atol=1e-5, rtol=0.)

    pred_actual = clf.fit(X, y).predict(X)
    pred_desired = cosine_similarity(freq, tf).argmax(axis=1)
    np.testing.assert_array_equal(pred_actual, pred_desired)



"""Code for discretizers."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

import numpy as np
from numba import njit, prange
from numba.typed import List
from scipy.stats import norm
from sklearn.base import BaseEstimator
from sklearn.utils.validation import check_array
from warnings import warn
from ..base import UnivariateTransformerMixin


@njit()
def _uniform_bins(sample_min, sample_max, n_samples, n_bins):
    bin_edges = np.empty((n_bins - 1, n_samples))
    for i in prange(n_samples):
        bin_edges[:, i] = np.linspace(
            sample_min[i], sample_max[i], n_bins + 1)[1:-1]
    return bin_edges


@njit()
def _digitize_1d(X, bins, n_samples, n_timestamps):
    X_digit = np.empty((n_samples, n_timestamps))
    for i in prange(n_samples):
        X_digit[i] = np.searchsorted(bins, X[i], side='left')
    return X_digit


@njit()
def _digitize_2d(X, bins, n_samples, n_timestamps):
    X_digit = np.empty((n_samples, n_timestamps))
    for i in prange(n_samples):
        X_digit[i] = np.searchsorted(bins[i], X[i], side='left')
    return X_digit


def _digitize(X, bins):
    n_samples, n_timestamps = X.shape
    if bins.ndim == 1:
        X_binned = _digitize_1d(X, bins, n_samples, n_timestamps)
    else:
        X_binned = _digitize_2d(X, bins, n_samples, n_timestamps)
    return X_binned.astype('int64')


@njit
def _reshape_with_nan(X, n_samples, lengths, max_length):
    X_fill = np.full((n_samples, max_length), np.nan)
    for i in prange(n_samples):
        X_fill[i, :lengths[i]] = X[i]
    return X_fill


class KBinsDiscretizer(BaseEstimator, UnivariateTransformerMixin):
    """Bin continuous data into intervals sample-wise.

    Parameters
    ----------
    n_bins : int (default = 5)
        The number of bins to produce. The intervals for the bins are
        determined by the minimum and maximum of the input data. It must
        be greater than or equal to 2.

    strategy : 'uniform', 'quantile' or 'normal' (default = 'quantile')
        Strategy used to define the widths of the bins:

        - 'uniform': All bins in each sample have identical widths
        - 'quantile': All bins in each sample have the same number of points
        - 'normal': Bin edges are quantiles from a standard normal distribution

    raise_warning : bool (default = True)
        If True, a warning is raised when the number of bins is smaller for
        at least one sample. In this case, you should consider decreasing the
        number of bins or removing these samples.

    Examples
    --------
    >>> from pyts.preprocessing import KBinsDiscretizer
    >>> X = [[0, 1, 0, 2, 3, 3, 2, 1],
    ...      [7, 0, 6, 1, 5, 3, 4, 2]]
    >>> discretizer = KBinsDiscretizer(n_bins=2)
    >>> print(discretizer.transform(X))
    [[0 0 0 1 1 1 1 0]
     [1 0 1 0 1 0 1 0]]

    """

    def __init__(self, n_bins=5, strategy='quantile', raise_warning=True):
        self.n_bins = n_bins
        self.strategy = strategy
        self.raise_warning = raise_warning

    def fit(self, X=None, y=None):
        """Pass.

        Parameters
        ----------
        X
            Ignored

        y
            Ignored

        Returns
        -------
        self : object

        """
        return self

    def transform(self, X):
        """Bin the data.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_timestamps)
            Data to transform.

        Returns
        -------
        X_new : array-like, shape = (n_samples, n_timestamps)
            Binned data.

        """
        X = check_array(X, dtype='float64')
        n_samples, n_timestamps = X.shape
        self._check_params(n_timestamps)

        bin_edges = self._compute_bins(
            X, n_samples, self.n_bins, self.strategy)
        X_new = _digitize(X, bin_edges)
        return X_new

    def _check_params(self, n_timestamps):
        if not isinstance(self.n_bins, (int, np.integer)):
            raise TypeError("'n_bins' must be an integer.")
        if not 2 <= self.n_bins:
            raise ValueError(
                "'n_bins' must be greater than or equal to 2 (got {0})."
                .format(self.n_bins)
            )
        if self.strategy not in ['uniform', 'quantile', 'normal']:
            raise ValueError("'strategy' must be either 'uniform', 'quantile' "
                             "or 'normal' (got {0}).".format(self.strategy))

    def _compute_bins(self, X, n_samples, n_bins, strategy):
        if strategy == 'normal':
            bin_edges = norm.ppf(np.linspace(0, 1, self.n_bins + 1)[1:-1])
        elif strategy == 'uniform':
            sample_min, sample_max = np.min(X, axis=1), np.max(X, axis=1)
            bin_edges = _uniform_bins(
                sample_min, sample_max, n_samples, n_bins).T
        else:
            bin_edges = np.percentile(
                X, np.linspace(0, 100, self.n_bins + 1)[1:-1], axis=1
            ).T
            mask = np.c_[
                ~np.isclose(0, np.diff(bin_edges, axis=1), rtol=0, atol=1e-8),
                np.full((n_samples, 1), True)
            ]
            if (self.n_bins > 2) and np.any(~mask):
                samples = np.where(np.any(~mask, axis=0))[0]
                if self.raise_warning:
                    warn("Some quantiles are equal. The number of bins will "
                         "be smaller for sample {0}. Consider decreasing the "
                         "number of bins or removing these samples."
                         .format(samples))
                lengths = np.sum(mask, axis=1)
                max_length = np.max(lengths)

                bin_edges_ = List()
                for i in range(n_samples):
                    bin_edges_.append(bin_edges[i][mask[i]])

                bin_edges = _reshape_with_nan(bin_edges_, n_samples,
                                              lengths, max_length)
        return bin_edges



"""Testing for k-nearest-neighbors."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

import numpy as np
import pytest
from pyts.classification import KNeighborsClassifier


X = np.arange(40).reshape(8, 5)
y = np.array([0, 0, 0, 1, 1, 0, 1, 1])
y_proba = np.vstack([1 - y, y]).T
X_test = X + 0.1


@pytest.mark.parametrize(
    'params',
    [({'metric': 'euclidean'}),
     ({'metric': 'manhattan'}),
     ({'metric': 'dtw'}),
     ({'metric': 'dtw_classic'}),
     ({'metric': 'dtw_sakoechiba'}),
     ({'metric': 'dtw_sakoechiba', 'metric_params': {}}),
     ({'metric': 'dtw_sakoechiba', 'metric_params': {'window_size': 0.5}}),
     ({'metric': 'dtw_itakura'}),
     ({'metric': 'dtw_itakura', 'metric_params': {}}),
     ({'metric': 'dtw_itakura', 'metric_params': {'max_slope': 3.}}),
     ({'metric': 'dtw_multiscale'}),
     ({'metric': 'dtw_fast'}),
     ({'metric': 'boss'})]
)
def test_actual_results(params):
    """Test that the actual results are the expected ones."""
    knn = KNeighborsClassifier(n_neighbors=1, **params)
    proba_actual = knn.fit(X, y).predict_proba(X_test)
    pred_actual = knn.predict(X_test)
    np.testing.assert_array_equal(proba_actual, y_proba)
    np.testing.assert_array_equal(pred_actual, y)
